# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
import json
import os
import pickle
import tkinter as tk
from tkinter import Button
from typing import Dict, List, Optional, Tuple, cast

import jsonpickle  # type: ignore
import jsons  # type: ignore
from sympy import Float, Symbol
from discopop_explorer.json_serializer import PatternBaseSerializer  # type: ignore
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.CostModels.DataTransfer.DataTransferCosts import add_data_transfer_costs
from discopop_library.discopop_optimizer.CostModels.utilities import get_performance_models_for_functions
from discopop_library.discopop_optimizer.DataTransfers.DataTransfers import calculate_data_transfers  # type: ignore
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.PETParser.PETParser import PETParser  # type: ignore

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.system.System import System
from discopop_library.discopop_optimizer.gui.presentation.OptionTable import (
    show_options,
    add_random_models,
)
from discopop_library.discopop_optimizer.gui.queries.ValueTableQuery import query_user_for_symbol_values
from discopop_library.discopop_optimizer.gui.widgets.ScrollableFrame import ScrollableFrameWidget
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from discopop_library.result_classes.DetectionResult import DetectionResult


def show_function_models(
    experiment: Experiment,
    parent_frame: tk.Frame,
    destroy_window_after_execution: bool,
    show_functions: Optional[List[FunctionRoot]] = None,
):
    considered_functions = show_functions if show_functions is not None else experiment.function_models
    # show function selection dialogue
    parent_frame.rowconfigure(0, weight=1)
    parent_frame.rowconfigure(1, weight=1)
    parent_frame.columnconfigure(1, weight=1)
    parent_frame.columnconfigure(0, weight=1)

    scrollable_frame_widget = ScrollableFrameWidget(parent_frame)
    scrollable_frame = scrollable_frame_widget.get_scrollable_frame()

    spawned_windows: List[tk.Toplevel] = []

    # populate scrollable frame
    for idx, function in enumerate(considered_functions):
        function_button = Button(
            scrollable_frame,
            text=function.name,
            command=lambda func=function: show_options(  # type: ignore
                # random options might be added by show_options
                experiment.detection_result.pet,
                experiment.optimization_graph,
                experiment,
                experiment.function_models[func],
                experiment.sorted_free_symbols,
                experiment.free_symbol_ranges,
                experiment.free_symbol_distributions,
                func,
                parent_frame,
                spawned_windows,
                window_title="Function: " + func.name,
            ),
        )
        function_button.grid(row=idx, column=0)

    # finalize scrollable frame
    scrollable_frame_widget.finalize(row_count=len(considered_functions), row=0, col=0)

    def __on_press():
        for w in spawned_windows:
            try:
                w.destroy()
            except tk.TclError:
                pass
        if destroy_window_after_execution:
            for c in parent_frame.winfo_children():
                c.destroy()
            parent_frame.winfo_toplevel().destroy()
        else:
            parent_frame.quit()

    # add exit button
    exit_button = Button(parent_frame, text="Exit", command=lambda: __on_press())  # type: ignore
    exit_button.grid(row=1, column=0)

    parent_frame.mainloop()


def perform_headless_execution(
    experiment: Experiment,
):
    print("Headless execution...")
    for function in experiment.function_models:
        print("\t", function.name)
        # generate random models
        updated_options = add_random_models(
            None,
            experiment.detection_result.pet,
            experiment.optimization_graph,
            experiment,
            experiment.function_models[function],
            experiment.sorted_free_symbols,
            experiment.free_symbol_ranges,
            experiment.free_symbol_distributions,
            function,
            None,
            [],
            show_results=False,
        )

        # save models
        experiment.function_models[function] = updated_options


def export_to_json(experiment: Experiment, export_path):
    # convert functionRoot in function_models to node ids
    to_be_added = []
    to_be_deleted = []
    for old_key in experiment.function_models:
        new_key = old_key.node_id
        to_be_added.append((new_key, experiment.function_models[old_key]))
        to_be_deleted.append(old_key)

    for k1 in to_be_deleted:
        del experiment.function_models[k1]
    for k2, v in to_be_added:
        experiment.function_models[k2] = v  # type: ignore

    experiment_dump_path: str = os.path.join(export_path, "last_experiment.pickle")
    if not os.path.exists(export_path):
        os.makedirs(export_path)
    pickle.dump(experiment, open(experiment_dump_path, "wb"))


def export_patterns_to_json(experiment: Experiment, export_path):
    detection_result_copy = copy.deepcopy(experiment.detection_result)
    # pet is not serializable and needs to be deleted
    del detection_result_copy.pet
    with open(export_path, "w+") as f:
        json.dump(detection_result_copy, f, indent=2, cls=PatternBaseSerializer)


def export_detection_result_to_json(experiment: Experiment, export_path):
    with open(export_path, "w+") as f:
        f.write(experiment.detection_result.dump_to_pickled_json())
        f.flush()
        f.close()


def restore_session(json_file: str) -> Experiment:
    experiment: Experiment = pickle.load(open(json_file, "rb"))

    # convert keys of function_models to FunctionRoot objects
    to_be_added = []
    to_be_deleted = []
    for old_key in experiment.function_models:
        new_key = cast(FunctionRoot, data_at(experiment.optimization_graph, cast(int, old_key)))
        to_be_added.append((new_key, experiment.function_models[old_key]))
        to_be_deleted.append(old_key)

    for k, v in to_be_added:
        experiment.function_models[k] = v
    for k in to_be_deleted:
        del experiment.function_models[k]

    return experiment


def create_optimization_graph(experiment: Experiment, arguments: OptimizerArguments):
    if arguments.verbose:
        print("Creating optimization graph...", end="")
    pet_parser = PETParser(experiment)
    experiment.optimization_graph, experiment.next_free_node_id = pet_parser.parse()
    if arguments.verbose:
        print("Done.")


def get_sequential_cost_model(experiment: Experiment) -> Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]]:
    # get performance models for sequential execution
    sequential_function_performance_models = get_performance_models_for_functions(
        experiment, experiment.optimization_graph
    )
    sequential_function_performance_models_with_transfers = calculate_data_transfers(
        experiment.optimization_graph, sequential_function_performance_models, experiment
    )
    sequential_complete_performance_models = add_data_transfer_costs(
        experiment.optimization_graph,
        sequential_function_performance_models_with_transfers,
        experiment,
    )
    return sequential_complete_performance_models


def initialize_free_symbol_ranges_and_distributions(
    experiment: Experiment, arguments: OptimizerArguments, system: System
):
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]] = dict()
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution] = dict()
    sorted_free_symbols = sorted(list(experiment.free_symbols), key=lambda x: x.name)
    symbol_values: List[
        Tuple[
            Symbol,
            Optional[float],
            Optional[float],
            Optional[float],
            Optional[FreeSymbolDistribution],
        ]
    ] = []

    if arguments.interactive:
        symbol_values = query_user_for_symbol_values(sorted_free_symbols, experiment.suggested_values)
    else:
        # use the suggested values
        for symbol in sorted_free_symbols:
            # check if symbol is already defined by the system
            found_symbol = False
            for entry in system.get_symbol_values_and_distributions():
                if entry[0] == symbol:
                    symbol_values.append(entry)
                    found_symbol = True
                    break
            if not found_symbol:
                # else, use a default value for non-system symbols
                symbol_values.append((symbol, experiment.suggested_values[symbol].evalf(), None, None, None))

    for symbol, value, start_value, end_value, symbol_distribution in symbol_values:
        if value is not None:
            experiment.substitutions[symbol] = Float(value)
        else:
            experiment.free_symbol_ranges[symbol] = (cast(float, start_value), cast(float, end_value))
            experiment.free_symbol_distributions[symbol] = cast(FreeSymbolDistribution, symbol_distribution)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
import subprocess
from typing import List, Tuple, Dict, Optional, Set

from discopop_library.LineMapping.load import load_line_mapping
from discopop_library.LineMapping.save import save_line_mapping


def apply_line_mapping_modifications_from_files(file_id: int, original_file: str, modified_file: str):
    """Calculates diff between original_file and modified_file and applied modifications from the diff to the line_mapping"""
    if not os.path.exists(original_file):
        raise FileNotFoundError(original_file)
    if not os.path.exists(modified_file):
        raise FileNotFoundError(original_file)

    command = [
        "diff",
        original_file,
        modified_file,
    ]
    result = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    if result.returncode == 1:
        # print("RESULT: ", result.returncode)
        # print("STDERR:")
        # print(result.stderr)
        # print("STDOUT: ")
        # print(result.stdout)
        diff = result.stdout
        # print("DIFF: ", diff)
    else:
        diff = ""

    apply_line_mapping_modifications_from_diff(file_id, diff)


def apply_line_mapping_modifications_from_diff(file_id: int, diff: str):
    """parse diff, apply line num modifications according to c,a,d values"""
    # get cleaned diff
    cleaned_diff: List[str] = []
    for line in diff.split("\n"):
        if line.startswith("<") or line.startswith(">") or line.startswith("-") or len(line) == 0:
            continue
        line = line.replace("\n", "")
        cleaned_diff.append(line)

    # get line_mapping
    line_mapping: Dict[str, Dict[str, int]] = load_line_mapping()

    deletions: Set[str] = set()
    shifts: Dict[str, int] = dict()

    # initialize
    for key in line_mapping[str(file_id)]:
        shifts[key] = 0

    # parse diff
    for diff_entry in cleaned_diff:
        if "a" in diff_entry:
            lhs = diff_entry[: diff_entry.index("a")]
            if "," in lhs:
                base_line = int(lhs.split(",")[0])
            else:
                base_line = int(lhs)
            rhs = diff_entry[diff_entry.index("a") + 1 :]
            if "," in rhs:
                added_lines_count = int(rhs.split(",")[1]) - int(rhs.split(",")[0]) + 1
            else:
                added_lines_count = 1
            for key in line_mapping[str(file_id)]:
                if line_mapping[str(file_id)][key] < 0:
                    # invalid due to deletion
                    continue
                if line_mapping[str(file_id)][key] > base_line:
                    shifts[key] += added_lines_count

        if "d" in diff_entry:
            lhs = diff_entry[: diff_entry.index("d")]
            if "," in lhs:
                base_line = int(lhs.split(",")[0])
                deleted_lines_count = int(lhs.split(",")[1]) - int(lhs.split(",")[0]) + 1
            else:
                base_line = int(lhs)
                deleted_lines_count = 1
            for key in line_mapping[str(file_id)]:
                if line_mapping[str(file_id)][key] < 0:
                    # invalid due to deletion
                    continue
                if line_mapping[str(file_id)][key] in range(base_line, base_line + deleted_lines_count):
                    deletions.add(key)
                elif line_mapping[str(file_id)][key] >= base_line + deleted_lines_count:
                    shifts[key] -= deleted_lines_count

        if "c" in diff_entry:
            lhs = diff_entry[: diff_entry.index("c")]
            if "," in lhs:
                base_line = int(lhs.split(",")[0])
                deleted_lines_count = int(lhs.split(",")[1]) - int(lhs.split(",")[0]) + 1
            else:
                base_line = int(lhs)
                deleted_lines_count = 1
            rhs = diff_entry[diff_entry.index("c") + 1 :]
            if "," in rhs:
                added_lines_count = int(rhs.split(",")[1]) - int(rhs.split(",")[0]) + 1
            else:
                added_lines_count = 1

            for key in line_mapping[str(file_id)]:
                if line_mapping[str(file_id)][key] < 0:
                    # invalid due to deletion
                    continue
                if line_mapping[str(file_id)][key] > base_line:
                    shifts[key] += added_lines_count

            for key in line_mapping[str(file_id)]:
                if line_mapping[str(file_id)][key] < 0:
                    # invalid due to deletion
                    continue
                if line_mapping[str(file_id)][key] in range(base_line, base_line + deleted_lines_count):
                    deletions.add(key)
                elif line_mapping[str(file_id)][key] >= base_line + deleted_lines_count:
                    shifts[key] -= deleted_lines_count

    # apply deletions
    for key in deletions:
        line_mapping[str(file_id)][key] = -1

    # apply shifts
    for key in shifts:
        if line_mapping[str(file_id)][key] < 0:
            # invalid due to deletion
            continue
        line_mapping[str(file_id)][key] = line_mapping[str(file_id)][key] + shifts[key]

    # save updated line mapping
    save_line_mapping(line_mapping)

    pass



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Tuple, Dict, Set, cast

from discopop_explorer.PEGraphX import EdgeType, DepType, PEGraphX, NodeID, CUNode, MemoryRegion
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)


def get_written_and_read_memory_regions_by_cu(
    contained_regions, pet: PEGraphX
) -> Tuple[Dict[NodeID, Set[MemoryRegion]], Dict[NodeID, Set[MemoryRegion]]]:
    all_function_cu_ids: Set[NodeID] = set()
    for region in contained_regions:
        parent_function = pet.get_parent_function(pet.node_at(region.node_id))

        subtree = pet.subtree_of_type(parent_function, CUNode)
        all_function_cu_ids.update([NodeID(n.id) for n in subtree])

    written_memory_regions_by_cu_id: Dict[NodeID, Set[MemoryRegion]] = dict()
    read_memory_regions_by_cu_id: Dict[NodeID, Set[MemoryRegion]] = dict()
    for cu_id in all_function_cu_ids:
        in_dep_edges = pet.in_edges(cu_id, EdgeType.DATA)
        out_dep_edges = pet.out_edges(cu_id, EdgeType.DATA)

        written_memory_regions = [
            MemoryRegion(cast(str, d.memory_region))
            for s, t, d in in_dep_edges
            if (d.dtype == DepType.RAW or d.dtype == DepType.WAW)
            and d.memory_region is not None
            and len(d.memory_region) != 0
        ]
        written_memory_regions += [
            MemoryRegion(cast(str, d.memory_region))
            for s, t, d in out_dep_edges
            if (d.dtype == DepType.WAR or d.dtype == DepType.WAW)
            and d.memory_region is not None
            and len(d.memory_region) != 0
        ]

        read_memory_regions = [
            MemoryRegion(cast(str, d.memory_region))
            for s, t, d in in_dep_edges
            if (d.dtype == DepType.WAR) and d.memory_region is not None and len(d.memory_region) != 0
        ]
        read_memory_regions += [
            MemoryRegion(cast(str, d.memory_region))
            for s, t, d in out_dep_edges
            if (d.dtype == DepType.RAW) and d.memory_region is not None and len(d.memory_region) != 0
        ]

        if cu_id not in written_memory_regions_by_cu_id:
            written_memory_regions_by_cu_id[cu_id] = set()
        written_memory_regions_by_cu_id[cu_id] = set(written_memory_regions)

        if cu_id not in read_memory_regions_by_cu_id:
            read_memory_regions_by_cu_id[cu_id] = set()
        read_memory_regions_by_cu_id[cu_id] = set(read_memory_regions)
    return written_memory_regions_by_cu_id, read_memory_regions_by_cu_id


def get_cu_and_varname_to_memory_regions(
    contained_regions, pet: PEGraphX, written_memory_regions_by_cu: Dict[NodeID, Set[MemoryRegion]]
) -> Dict[NodeID, Dict[VarName, Set[MemoryRegion]]]:
    # dict -> {Cu_ID: {var_name: [memory regions]}}
    result_dict: Dict[NodeID, Dict[VarName, Set[MemoryRegion]]] = dict()

    all_function_cu_ids: Set[NodeID] = set()
    for region in contained_regions:
        parent_function = pet.get_parent_function(pet.node_at(region.node_id))

        subtree = pet.subtree_of_type(parent_function, CUNode)
        all_function_cu_ids.update([NodeID(n.id) for n in subtree])

    for cu_id in all_function_cu_ids:
        if cu_id not in result_dict:
            result_dict[cu_id] = dict()

        # only out_deps considered, as in_deps might use variable names
        # which originate from different source code scopes
        out_dep_edges = pet.out_edges(cu_id, EdgeType.DATA)
        for _, _, dep in out_dep_edges:
            if dep.var_name is None or dep.memory_region is None or len(dep.memory_region) == 0:
                continue
            if dep.var_name not in result_dict[cu_id]:
                result_dict[cu_id][VarName(dep.var_name)] = set()
            result_dict[cu_id][VarName(dep.var_name)].add(dep.memory_region)

    return result_dict


def get_memory_region_to_cu_and_variables_dict(
    cu_and_variable_to_memory_regions: Dict[NodeID, Dict[VarName, Set[MemoryRegion]]]
) -> Dict[MemoryRegion, Dict[NodeID, Set[VarName]]]:
    # inverts the given cu_and_variable_to_memory_regions dictionary
    result_dict: Dict[MemoryRegion, Dict[NodeID, Set[VarName]]] = dict()

    for cu_id in cu_and_variable_to_memory_regions:
        for var_name in cu_and_variable_to_memory_regions[cu_id]:
            for mem_reg in cu_and_variable_to_memory_regions[cu_id][var_name]:
                if mem_reg not in result_dict:
                    result_dict[mem_reg] = dict()
                if cu_id not in result_dict[mem_reg]:
                    result_dict[mem_reg][cu_id] = set()
                result_dict[mem_reg][cu_id].add(var_name)
    return result_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
from typing import Any, Dict, Optional, cast
from discopop_explorer.PEGraphX import NodeID, PEGraphX
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID
from discopop_library.discopop_optimizer.classes.types.DataAccessType import (
    WriteDataAccess,
    write_data_access_from_dict,
)


class Update(object):
    source_node_id: int
    target_node_id: int
    source_device_id: DeviceID
    target_device_id: DeviceID
    write_data_access: WriteDataAccess
    is_first_data_occurrence: bool
    source_cu_id: Optional[NodeID]
    target_cu_id: Optional[NodeID]

    def __init__(
        self,
        source_node_id: int,
        target_node_id: int,
        source_device_id: DeviceID,
        target_device_id: DeviceID,
        write_data_access: WriteDataAccess,
        is_first_data_occurrence: bool,
        source_cu_id: Optional[NodeID],
        target_cu_id: Optional[NodeID],
    ):
        self.source_node_id = source_node_id
        self.target_node_id = target_node_id
        self.source_device_id = source_device_id
        self.target_device_id = target_device_id
        self.write_data_access = write_data_access
        self.is_first_data_occurrence = is_first_data_occurrence
        self.source_cu_id = source_cu_id
        self.target_cu_id = target_cu_id

    def __str__(self):
        result_str = "First" if self.is_first_data_occurrence else ""
        return (
            result_str
            + "Update("
            + str(self.source_node_id)
            + "@"
            + str(self.source_device_id)
            + ", "
            + str(self.target_node_id)
            + "@"
            + str(self.target_device_id)
            + ", "
            + " : "
            + str(self.write_data_access.memory_region)
            + " --> "
            + str(self.write_data_access.var_name)
            + ")"
        )

    def get_pattern_string(self, pet: PEGraphX) -> str:
        result_dict: Dict[str, Any] = dict()
        result_dict["source_device_id"] = self.source_device_id
        result_dict["target_device_id"] = self.target_device_id
        result_dict["openmp_source_device_id"] = self.source_device_id
        result_dict["openmp_target_device_id"] = self.target_device_id
        result_dict["is_first_data_occurrence"] = self.is_first_data_occurrence
        result_dict["var_name"] = self.write_data_access.var_name
        result_dict["start_line"] = pet.node_at(cast(NodeID, self.source_cu_id)).start_position()
        result_dict["end_line"] = pet.node_at(cast(NodeID, self.target_cu_id)).start_position()
        return json.dumps(result_dict)

    def toDict(self) -> Dict[str, Any]:
        result_dict: Dict[str, Any] = {}
        result_dict["source_node_id"] = self.source_node_id
        result_dict["target_node_id"] = self.target_node_id
        result_dict["source_device_id"] = self.source_device_id
        result_dict["target_device_id"] = self.target_device_id
        result_dict["is_first_data_occurrence"] = self.is_first_data_occurrence
        result_dict["write_data_access"] = self.write_data_access.toDict()
        result_dict["source_cu_id"] = self.source_cu_id
        result_dict["target_cu_id"] = self.target_cu_id
        return result_dict


def construct_update_from_dict(values: Dict[str, Any]) -> Update:
    update = Update(
        values["source_node_id"],
        values["target_node_id"],
        values["source_device_id"],
        values["target_device_id"],
        write_data_access_from_dict(values["write_data_access"]),
        values["is_first_data_occurrence"],
        values["source_cu_id"],
        values["target_cu_id"],
    )
    return update



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os.path
from pathlib import Path


def get_version() -> str:
    parent_folder_path = Path(__file__).parent.absolute()
    with open(os.path.join(parent_folder_path, "VERSION")) as f:
        return f.read().rstrip()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast, TextIO

from .PEGraphX import (
    CUNode,
    FunctionNode,
    LineID,
    LoopNode,
    NodeID,
    PEGraphX,
    Node,
    DepType,
    EdgeType,
)
from .parser import parse_inputs


def __collect_children_ids(pet: PEGraphX, parent_id: NodeID, children_ids: List[NodeID]):
    if parent_id in children_ids:
        # this id has already been processed. No need to go through it again
        return children_ids
    children_ids.append(parent_id)
    children_ids = list(set(children_ids))
    # collect all of its children
    for child_node in pet.direct_children_or_called_nodes(pet.node_at(parent_id)):
        children_ids += __collect_children_ids(pet, child_node.id, children_ids)
        children_ids = list(set(children_ids))
    return children_ids


def __recursive_call_inside_loop(pet: PEGraphX, recursive_function_call: str) -> bool:
    """checks if the given recursive function call occurs in any loop body.
    :param pet: PET Graph
    :param recursive_function_call: string representation of a recursive function call, extracted from cu-xml
    :return: True, if recursive call inside any loop body. False otherwise."""
    for tmp_cu in pet.all_nodes(LoopNode):
        if __line_contained_in_region(
            LineID(recursive_function_call.split(" ")[-1].replace(",", "")),
            tmp_cu.start_position(),
            tmp_cu.end_position(),
        ):
            return True
    return False


def __recursive_function_called_multiple_times_inside_function(pet: PEGraphX, recursive_function_call: str) -> bool:
    """checks if the given recursive function is called multiple times from within a functions body
    :param pet: PET Graph
    :param recursive_function_call: string representation of a recursive function call, extracted from cu-xml
    :return: True, if multiple calls exists. False otherwise."""
    for tmp_func_cu in pet.all_nodes(FunctionNode):
        # 1. get parent function of recursive function call
        if not __line_contained_in_region(
            LineID(recursive_function_call.split(" ")[-1].replace(",", "")),
            tmp_func_cu.start_position(),
            tmp_func_cu.end_position(),
        ):
            continue
        # recursive function call contained in tmp_func_cu
        # 2. check if multiple calls to recursive function exist in tmp_func_cus body
        # by listing cu nodes in function body.
        # get cu's inside function by traversing child edges
        queue: List[Node] = [tmp_func_cu]
        contained_cus: List[Node] = []
        while len(queue) > 0:
            cur_cu = queue.pop(0)
            if __line_contained_in_region(
                cur_cu.start_position(), tmp_func_cu.start_position(), tmp_func_cu.end_position()
            ) and __line_contained_in_region(
                cur_cu.end_position(), tmp_func_cu.start_position(), tmp_func_cu.end_position()
            ):
                # cur_cu contained in tmp_func_cu's scope
                if cur_cu not in contained_cus:
                    contained_cus.append(cur_cu)
                # append cur_cu children to queue
                for child_edge in pet.out_edges(cur_cu.id, [EdgeType.CHILD, EdgeType.CALLSNODE]):
                    child_cu = pet.node_at(child_edge[1])
                    if child_cu not in queue and child_cu not in contained_cus:
                        queue.append(child_cu)
        # get recursive function calls from contained_cus
        rec_calls: List[str] = []
        for tmp_cu in contained_cus:
            rec_calls += tmp_cu.recursive_function_calls
        # remove None and prune rec_calls to called function names only
        rec_calls = [e for e in rec_calls if e is not None]
        rec_calls = [e.split(" ")[0] for e in rec_calls]
        # check if recursive_function_call occurs at least twice
        if rec_calls.count(recursive_function_call.split(" ")[0]) > 1:
            return True
    return False


def __output_dependencies_of_type(
    pet: PEGraphX,
    child_id: NodeID,
    children_ids: List[NodeID],
    output_file: TextIO,
    dep_type: DepType,
    dep_identifier: str,
):
    """check for and output dependencies of the given type
    :param pet: PET Graph
    :param child_id: specific node id, taken from children_ids
    :param children_ids: list of children node id's of a given node.
    :param output_file: file to be written
    :param dep_type: type of dependency to be handled
    :param dep_identifier: identifier corresponding to the given dep_type (|RAW|, |WAR|, |WAW|)
    """
    for dep in pet.in_edges(child_id, EdgeType.DATA):
        if dep[2].dtype is not dep_type:
            continue
        if dep[2].source_line is None or dep[2].var_name is None or dep[2].sink_line is None:
            continue
        # check if the CUid of the dep exists in children_ids
        if dep[0] in children_ids:
            # CUid found in children_ids. So print the line numbers of this dependence
            output_file.write(
                cast(str, dep[2].sink_line)
                + dep_identifier
                + cast(str, dep[2].source_line)
                + "|"
                + dep[2].var_name
                + ","
            )


def __search_recursive_calls(pet: PEGraphX, output_file, node: Node):
    if not isinstance(node, CUNode):
        return
    for recursive_function_call in node.recursive_function_calls:
        if recursive_function_call is None:
            continue
        # check if recursive function call occurs inside loop (check if line contained in lines of any loop cu)
        contained_in_loop = __recursive_call_inside_loop(pet, recursive_function_call)
        # check if recursive function call is called multiple times
        called_multiple_times = __recursive_function_called_multiple_times_inside_function(pet, recursive_function_call)

        # check if recursive function is called inside loop or multiple times
        if not (contained_in_loop or called_multiple_times):
            continue

        output_file.write(recursive_function_call + " ")

        children_ids: List[NodeID] = []
        # Output RAW deps. First collect all children IDs of currently called
        # function. Then go through the CU type nodes in them and record all deps
        # that are on the cus in this list.
        children_ids = __collect_children_ids(pet, node.id, children_ids)

        for child_id in children_ids:
            # node type is not cu so goto next node
            if not isinstance(pet.node_at(child_id), CUNode):
                continue
            __output_dependencies_of_type(pet, child_id, children_ids, output_file, DepType.RAW, "|RAW|")
            __output_dependencies_of_type(pet, child_id, children_ids, output_file, DepType.WAR, "|WAR|")
            __output_dependencies_of_type(pet, child_id, children_ids, output_file, DepType.WAW, "|WAW|")

        output_file.write("\n")


def cu_instantiation_input_cpp(pet: PEGraphX, output_file: str):
    """translation of CUInstantiationInput.cpp, previously contained in discopop-analyzer/analyzer/src.
    Wrapper to gather information on recursive function calls for CU Instantiation.
    :param pet: PET Graph
    :param output_file: Path to storage location of generated Data_CUInst.txt"""
    with open(output_file, "w+") as data_cu_inst_file:
        for node in pet.all_nodes():
            __search_recursive_calls(pet, data_cu_inst_file, node)


def wrapper(cu_xml, dep_file, loop_counter_file, reduction_file, output_file):
    """Wrapper to generate the Data_CUInst.txt file, required for the generation of CUInstResult.txt"""
    # 1. generate PET Graph
    pet = PEGraphX.from_parsed_input(*parse_inputs(cu_xml, dep_file, loop_counter_file, reduction_file))
    # 2. Generate Data_CUInst.txt
    cu_instantiation_input_cpp(pet, output_file)


def __line_contained_in_region(test_line: LineID, start_line: LineID, end_line: LineID) -> bool:
    """check if test_line is contained in [startLine, endLine].
    Return True if so. False else.
    :param test_line: <fileID>:<line>
    :param start_line: <fileID>:<line>
    :param end_line: <fileID>:<line>
    :return: bool
    """
    test_line_file_id = int(test_line.split(":")[0])
    test_line_line = int(test_line.split(":")[1])
    start_line_file_id = int(start_line.split(":")[0])
    start_line_line = int(start_line.split(":")[1])
    end_line_file_id = int(end_line.split(":")[0])
    end_line_line = int(end_line.split(":")[1])
    if (
        test_line_file_id == start_line_file_id == end_line_file_id
        and start_line_line <= test_line_line <= end_line_line
    ):
        return True
    return False



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, List, Tuple, Union

from .Microbench import Microbench
from .Microbench import MicrobenchCoordinate, MicrobenchDimension, MicrobenchType


# This class can be used to mix two microbench models.
# The inner model will be used for evaluation of the lower values
# The outer model will be used for evaluation of the higher values (if any coordinate point is higher than the threshold)
class MixedMicrobench(Microbench):
    inner: Microbench
    outer: Microbench
    threshold: MicrobenchCoordinate

    def __init__(self, inner: Microbench, outer: Microbench, threshold: MicrobenchCoordinate):
        self.inner = inner
        self.outer = outer
        self.threshold = threshold

    def getMeasurements(
        self,
    ) -> Dict[MicrobenchType, Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],]:
        raise TypeError(
            "This MixedMicrobench might be based on two different sets of measurements. Use getInnerMeasurements() or getOuterMeasurements() instead."
        )

    def getInnerMeasurements(
        self,
    ) -> Dict[MicrobenchType, Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],]:
        return self.inner.getMeasurements()

    def getOuterMeasurements(
        self,
    ) -> Dict[MicrobenchType, Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],]:
        return self.outer.getMeasurements()

    def toJSON(self) -> str:
        raise TypeError("This class does not support JSON serialization.")

    def evaluateInterpolation(
        self,
        benchType: MicrobenchType,
        benchDim: MicrobenchDimension,
        benchCoord: Union[MicrobenchCoordinate, Tuple[int, float, float]],
    ) -> float:
        if benchCoord[1:] <= self.threshold[1:]:
            return self.inner.evaluateInterpolation(benchType, benchDim, benchCoord)
        else:
            return self.outer.evaluateInterpolation(benchType, benchDim, benchCoord)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Dict, List, Any, Optional

from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import (
    EntryPointType,
    EntryPointPositioning,
    UpdateType,
    ExitPointType,
    ExitPointPositioning,
)
from discopop_library.CodeGenerator.classes.Enums import (
    OmpConstructPositioning,
    PragmaPosition,
)
from discopop_library.CodeGenerator.classes.Pragma import Pragma
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID


class UnpackedSuggestion(object):
    type: str
    values: Dict[str, Any]
    file_id: int
    start_line: int
    end_line: int
    device_id: Optional[DeviceID]
    device_type: Optional[DeviceTypeEnum]
    host_device_id: int

    def __init__(self, type_str: str, values: Dict[str, Any], device_id=None, device_type=None, host_device_id=0):
        self.type = type_str
        self.values = values
        self.device_id = device_id
        self.device_type = device_type
        self.host_device_id = host_device_id

        # get start and end line of target section
        self.file_id = int(self.values["start_line"].split(":")[0])
        self.start_line = int(self.values["start_line"].split(":")[1])
        self.end_line = int(self.values["end_line"].split(":")[1])

        # read device id and type if present
        if self.device_id is None and "device_id" in self.values:
            if self.values["device_id"] is not None:
                self.device_id = int(self.values["device_id"])
        if self.device_type is None and "device_type" in self.values:
            if self.values["device_type"] is not None:
                self.device_type = self.values["device_type"]

    def __get_device_update_pragmas(self):
        pragmas = []
        pragma = Pragma()

        source_device_id = int(self.values["source_device_id"])
        target_device_id = int(self.values["target_device_id"])
        var_name = self.values["var_name"]
        is_first_data_occurrence: bool = self.values["is_first_data_occurrence"]
        openmp_source_device_id = self.values["openmp_source_device_id"]
        openmp_target_device_id = self.values["openmp_target_device_id"]
        print("IS FIRST DATA OCCURRENCE?: ", is_first_data_occurrence)

        if source_device_id == self.host_device_id and target_device_id == self.host_device_id:
            # no update required
            pass
        elif source_device_id == self.host_device_id and target_device_id != self.host_device_id:
            # update type to
            if is_first_data_occurrence:
                pragma.pragma_str = "#pragma omp target enter data map(to:"
            else:
                pragma.pragma_str = "#pragma omp target update to("
            pragma.pragma_str += (
                var_name
                + ") device("
                # + str(openmp_source_device_id)
                # + " -> "
                + str(openmp_target_device_id)
                + ")"
            )

        elif source_device_id != self.host_device_id and target_device_id == self.host_device_id:
            # update type from
            if is_first_data_occurrence:
                pragma.pragma_str = "#pragma omp target exit data map(from:"
            else:
                pragma.pragma_str = "#pragma omp target update from("

            pragma.pragma_str += (
                var_name
                + ") device("
                + str(openmp_source_device_id)
                # + " -> "
                # + str(openmp_target_device_id)
                + ")"
            )

        else:
            # todo Implement updates between devices
            #  Suggest two update instructions instead of one, or make sure that openMP supports device-device updates

            # raise NotImplementedError("Updates between devices are not yet implemented!")
            # update between two devices
            pragma.pragma_str = (
                "#pragma omp target update Device("
                + str(source_device_id)
                + ") --> Device("
                + str(openmp_source_device_id)
                + " -> "
                + str(openmp_target_device_id)
                + ") Var("
                + var_name
                + ")"
            )

        #            #  map to host
        #            pragma.pragma_str = (
        #                # "#pragma omp target map(from:"
        #                    "#pragma omp target enter data map(to:"
        #                    + var_name
        #                    + ") device("
        #                    + str(source_device_id)
        #                    + ")"
        #            )
        #            #  map from host to second device
        #            pragma.pragma_str = (
        #                    "#pragma omp target data map(from:"
        #                    + var_name
        #                    + ") device("
        #                    + str(target_device_id)
        #                    + ")"
        #            )

        pragma.file_id = self.file_id
        pragma.start_line = self.start_line
        pragma.end_line = self.end_line
        pragma.pragma_position = PragmaPosition.BEFORE_END

        if len(pragma.pragma_str) != 0:
            pragmas.append(pragma)
        return pragmas

    def __get_do_all_and_reduction_pragmas(self, is_gpu_pragma: bool) -> List[Pragma]:
        pragmas = []
        pragma = Pragma()
        if is_gpu_pragma:
            # get device id for execution (else branch for compatibility with "legacy" gpu suggestions)
            if self.device_id is not None:
                device_id = self.device_id
            else:
                device_id = self.values["dp_optimizer_device_id"]

            pragma.pragma_str = "#pragma omp target teams distribute parallel for "
            pragma.pragma_str += "device(" + str(device_id) + ") "
        else:
            pragma.pragma_str = "#pragma omp parallel for "
        if "collapse_level" in self.values:
            if self.values["collapse_level"] > 1:
                pragma.pragma_str += "collapse(" + str(self.values["collapse_level"]) + ") "
        if len(self.values["first_private"]) > 0:
            pragma.pragma_str += "firstprivate(" + ",".join(self.values["first_private"]) + ") "
        if len(self.values["private"]) > 0:
            pragma.pragma_str += "private(" + ",".join(self.values["private"]) + ") "
        if len(self.values["last_private"]) > 0:
            pragma.pragma_str += "lastprivate(" + ",".join(self.values["last_private"]) + ") "
        if len(self.values["shared"]) > 0:
            pragma.pragma_str += "shared(" + ",".join(self.values["shared"]) + ") "
        if len(self.values["reduction"]) > 0:
            reductions_dict: Dict[str, List[str]] = dict()
            for entry in self.values["reduction"]:
                red_type: str = entry.split(":")[0]
                var: str = entry.split(":")[1]
                if red_type not in reductions_dict:
                    reductions_dict[red_type] = []
                reductions_dict[red_type].append(var)
            for red_type in reductions_dict:
                pragma.pragma_str += "reduction(" + red_type + ":" + ",".join(reductions_dict[red_type]) + ") "

        pragma.file_id = self.file_id
        pragma.start_line = self.start_line
        pragma.end_line = self.end_line

        pragmas.append(pragma)
        return pragmas

    def __get_pipeline_pragmas(self) -> List[Pragma]:
        pragmas = []

        for stage in self.values["stages"]:
            pragma = Pragma()
            pragma.file_id = self.file_id
            pragma.start_line = int(stage["startsAtLine"].split(":")[1])
            pragma.end_line = int(stage["endsAtLine"].split(":")[1])
            pragma.pragma_str = "#pragma omp task "
            if len(stage["first_private"]) > 0:
                pragma.pragma_str += "firstprivate(" + ",".join(stage["first_private"]) + ") "
            if len(stage["private"]) > 0:
                pragma.pragma_str += "private(" + ",".join(stage["private"]) + ") "
            if len(stage["shared"]) > 0:
                pragma.pragma_str += "shared(" + ",".join(stage["shared"]) + ") "
            if len(stage["reduction"]) > 0:
                reductions_dict: Dict[str, List[str]] = dict()
                for entry in stage["reduction"]:
                    red_type: str = entry.split(":")[0]
                    var: str = entry.split(":")[1]
                    if red_type not in reductions_dict:
                        reductions_dict[red_type] = []
                    reductions_dict[red_type].append(var)
                for red_type in reductions_dict:
                    pragma.pragma_str += "reduction(" + red_type + ":" + ",".join(reductions_dict[red_type]) + ") "
            if len(stage["in_deps"]) > 0:
                pragma.pragma_str += "depends(in:" + ",".join(stage["in_deps"]) + ") "
            if len(stage["out_deps"]) > 0:
                pragma.pragma_str += "depends(out:" + ",".join(stage["out_deps"]) + ") "
            if len(stage["in_out_deps"]) > 0:
                pragma.pragma_str += "depends(inout:" + ",".join(stage["in_out_deps"]) + ") "
            pragmas.append(pragma)
        return pragmas

    def __get_simple_gpu_pragmas(
        self,
        region_start,
        region_end,
        contained_loops,
        map_to_vars,
        map_from_vars,
        map_to_from_vars,
        map_alloc_vars,
        map_delete_vars,
        consumed_vars,
        produced_vars,
        indentation: int = 0,
        ignore_mapping_clauses: bool = False,
    ) -> List[Pragma]:
        pragmas = []

        region_start_line = int(region_start.split(":")[1])
        region_end_line = int(region_end.split(":")[1])
        for loop in contained_loops:
            loop_start = loop["start_line"]
            loop_start_line = int(loop_start.split(":")[1])
            loop_end = loop["end_line"]
            loop_end_line = int(loop_end.split(":")[1])
            loop_pragma = Pragma()
            loop_pragma.start_line = loop_start_line
            loop_pragma.end_line = loop_end_line
            loop_pragma.file_id = int(loop_start.split(":")[0])
            for construct in loop["constructs"]:
                construct_start = construct["line"]
                mark_invalid = False
                if ":" not in construct_start:
                    # invalid construct, line has no position
                    mark_invalid = True
                    construct_start = loop_start
                construct_start_line = int(construct_start.split(":")[1])
                child_pragma = Pragma()
                if mark_invalid:
                    child_pragma.pragma_str = "// INVALID - MISSING POSITION:: " + construct["name"] + " "
                else:
                    child_pragma.pragma_str = construct["name"] + " "
                if loop["collapse"] > 1:
                    child_pragma.pragma_str += "collapse(" + str(loop["collapse"]) + ") "
                for clause in construct["clauses"]:
                    if ignore_mapping_clauses:
                        if clause.startswith("map("):
                            continue
                    child_pragma.pragma_str += clause + " "
                # determine start_line and end_line
                if construct_start_line == loop_start_line:
                    # if construct targets loop, use loop scope
                    start_line = loop_start_line
                    end_line = loop_end_line
                elif loop_start_line <= construct_start_line <= loop_end_line:
                    # if construct is inside loop scope, start at construct line and end at loop scope
                    # (should not be used currently)
                    start_line = construct_start_line
                    end_line = loop_end_line
                else:
                    # else, use construct line
                    start_line = construct_start_line
                    end_line = construct_start_line
                # determine positioning of the pragma
                if construct["positioning"] == OmpConstructPositioning.BEFORE_LINE:
                    child_pragma.pragma_position = PragmaPosition.BEFORE_START
                elif construct["positioning"] == OmpConstructPositioning.AFTER_LINE:
                    child_pragma.pragma_position = PragmaPosition.AFTER_START
                else:
                    raise ValueError("Unsupported positioning information: ", construct["positioning"])
                # create pragma for visualization
                child_pragma.start_line = start_line
                child_pragma.end_line = end_line
                child_pragma.file_id = self.file_id
                loop_pragma.children.append(child_pragma)
            pragmas.append(loop_pragma)
        return pragmas

    def __get_update_pragmas(self, update_instructions) -> List[Pragma]:
        pragmas = []
        for source_cu_id, sink_cu_id, update_type, target_var, pragma_line in update_instructions:
            pragma = Pragma()
            pragma.pragma_str = "#pragma omp target update "

            if update_type == UpdateType.TO_DEVICE:
                # to device means host is writing, so update after the instruction
                pragma.pragma_position = PragmaPosition.BEFORE_START  # PragmaPosition.AFTER_END
                pragma.pragma_str += "to("
            elif update_type == UpdateType.FROM_DEVICE:
                # from device means host is reading, so update before the instruction
                pragma.pragma_position = PragmaPosition.BEFORE_START
                pragma.pragma_str += "from("
            elif update_type == UpdateType.ALLOCATE:
                # allocate memory, not written to before
                pragma.pragma_position = PragmaPosition.BEFORE_START
                pragma.pragma_str += "alloc("
            else:
                raise ValueError("Unsupported update type: ", update_type)
            pragma.pragma_str += target_var + ") "
            pragma.pragma_str += " // @CU " + source_cu_id + " -> " + sink_cu_id + " "
            pragma_line_num = int(pragma_line.split(":")[1])
            pragma.start_line = pragma_line_num
            pragma.end_line = pragma_line_num
            pragma.file_id = self.file_id
            pragmas.append(pragma)
        return pragmas

    def __get_data_region_dependencies(self, depend_in, depend_out) -> List[Pragma]:
        pragmas = []
        for var_name, cu_id, pragma_line in depend_in:
            pragma = Pragma()
            pragma.pragma_str = "#depend in(" + var_name + ") @ " + cu_id + " "
            pragma_line_num = int(pragma_line.split(":")[1])
            pragma.start_line = pragma_line_num
            pragma.end_line = pragma_line_num
            pragma.file_id = self.file_id
            pragma.pragma_position = PragmaPosition.BEFORE_START
            pragmas.append(pragma)

        for var_name, cu_id, pragma_line in depend_out:
            pragma = Pragma()
            pragma.pragma_str = "#depend out(" + var_name + ") @ " + cu_id + " "
            pragma_line_num = int(pragma_line.split(":")[1])
            pragma.start_line = pragma_line_num
            pragma.end_line = pragma_line_num
            pragma.file_id = self.file_id
            pragma.pragma_position = PragmaPosition.BEFORE_START
            pragmas.append(pragma)

        return pragmas

    def __get_data_region_pragmas(self, entry_points, exit_points) -> List[Pragma]:
        pragmas = []
        for (
            var_name,
            source_cu_id,
            sink_cu_id,
            entry_point_type,
            pragma_line,
            entry_point_positioning,
        ) in entry_points:
            pragma = Pragma()
            pragma.pragma_str = "#pragma omp target enter data "
            if entry_point_type == EntryPointType.TO_DEVICE:
                pragma.pragma_str += "map(to: "
            elif entry_point_type == EntryPointType.ALLOCATE:
                pragma.pragma_str += "map(alloc: "
            elif entry_point_type == EntryPointType.ASYNC_TO_DEVICE:
                pragma.pragma_str += "nowait map(to: "
            elif entry_point_type == EntryPointType.ASYNC_ALLOCATE:
                pragma.pragma_str += "nowait map(alloc:"
            else:
                raise ValueError("Usupported EntryPointType: ", entry_point_type)
            pragma.pragma_str += var_name + ") "
            pragma.pragma_str += " // @CU " + source_cu_id + " -> " + sink_cu_id + " "
            pragma_line_num = int(pragma_line.split(":")[1])
            pragma.start_line = pragma_line_num
            pragma.end_line = pragma_line_num
            pragma.file_id = self.file_id
            if entry_point_positioning == EntryPointPositioning.BEFORE_CU:
                pragma.pragma_position = PragmaPosition.BEFORE_START
            elif entry_point_positioning == EntryPointPositioning.AFTER_CU:
                pragma.pragma_position = PragmaPosition.AFTER_END
            else:
                raise ValueError("Usupported ExitPointPositioning: ", entry_point_positioning)
            pragmas.append(pragma)

        for (
            var_name,
            source_cu_id,
            sink_cu_id,
            exit_point_type,
            pragma_line,
            exit_point_positioning,
        ) in exit_points:
            pragma = Pragma()
            pragma.pragma_str = "#pragma omp target exit data "
            if exit_point_type == ExitPointType.FROM_DEVICE:
                pragma.pragma_str += "map(from: "
            elif exit_point_type == ExitPointType.DELETE:
                pragma.pragma_str += "map(delete: "
            elif exit_point_type == ExitPointType.ASYNC_FROM_DEVICE:
                pragma.pragma_str += "nowait map(from: "
            else:
                raise ValueError("Usupported ExitPointType: ", exit_point_type)
            pragma.pragma_str += var_name + ") "
            pragma.pragma_str += " // @CU " + source_cu_id + " -> " + sink_cu_id + " "
            pragma_line_num = int(pragma_line.split(":")[1])
            pragma.start_line = pragma_line_num
            pragma.end_line = pragma_line_num
            pragma.file_id = self.file_id
            if exit_point_positioning == ExitPointPositioning.BEFORE_CU:
                pragma.pragma_position = PragmaPosition.BEFORE_START
            elif exit_point_positioning == ExitPointPositioning.AFTER_CU:
                pragma.pragma_position = PragmaPosition.AFTER_END
            else:
                raise ValueError("Usupported ExitPointPositioning: ", exit_point_positioning)
            pragmas.append(pragma)
        return pragmas

    def __get_combined_gpu_pragmas(self) -> List[Pragma]:
        pragmas = []
        # add async data movement
        pragmas += self.__get_data_region_pragmas(
            self.values["data_region_entry_points"], self.values["data_region_exit_points"]
        )
        # add dependencies
        pragmas += self.__get_data_region_dependencies(
            self.values["data_region_depend_in"], self.values["data_region_depend_out"]
        )

        # add update instructions to pragmas
        pragmas += self.__get_update_pragmas(self.values["update_instructions"])

        # add gpu loops
        for region in self.values["contained_regions"]:
            pragmas += self.__get_simple_gpu_pragmas(
                region["start_line"],
                region["end_line"],
                region["contained_loops"],
                region["map_to_vars"],
                region["map_from_vars"],
                region["map_to_from_vars"],
                region["map_alloc_vars"],
                region["map_delete_vars"],
                region["consumed_vars"],
                region["produced_vars"],
                indentation=0,
                ignore_mapping_clauses=True,
            )

        return pragmas

    def get_pragmas(self) -> List[Pragma]:
        """returns a list of source code lines and pragmas to be inserted into the code preview"""
        pragmas = []

        if self.type in ["do_all", "reduction", "gpu_do_all", "gpu_reduction"]:
            execute_on_gpu = "gpu_" in self.type or self.device_type == DeviceTypeEnum.GPU
            pragmas += self.__get_do_all_and_reduction_pragmas(execute_on_gpu)
            return pragmas
        elif self.type == "pipeline":
            pragmas += self.__get_pipeline_pragmas()
        elif self.type == "simple_gpu":
            pragmas += self.__get_simple_gpu_pragmas(
                self.values["start_line"],
                self.values["end_line"],
                self.values["contained_loops"],
                self.values["map_to_vars"],
                self.values["map_from_vars"],
                self.values["map_to_from_vars"],
                self.values["map_alloc_vars"],
                self.values["map_delete_vars"],
                self.values["consumed_vars"],
                self.values["produced_vars"],
            )
        elif self.type == "combined_gpu":
            pragmas += self.__get_combined_gpu_pragmas()
        elif self.type == "device_update":
            pragmas += self.__get_device_update_pragmas()
        else:
            pragma = Pragma()
            pragma.file_id = self.file_id
            pragma.start_line = self.start_line
            pragma.end_line = self.end_line
            pragma.pragma_str = "#CURRENTLY UNSUPPORTED PREVIEW FOR TYPE: " + self.type
            pragmas.append(pragma)
        return pragmas



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Any, Dict, Optional

from discopop_explorer.PEGraphX import MemoryRegion


class ReadDataAccess(object):
    memory_region: MemoryRegion
    var_name: Optional[str]

    def __init__(self, memory_region: MemoryRegion, var_name: Optional[str]):
        self.memory_region = memory_region
        self.var_name = var_name

    def __str__(self):
        return "R(" + self.memory_region + ")"


class WriteDataAccess(object):
    memory_region: MemoryRegion
    unique_id: int
    var_name: Optional[str]

    def __init__(self, memory_region: MemoryRegion, unique_id: int, var_name: Optional[str]):
        self.memory_region = memory_region
        self.unique_id = unique_id
        self.var_name = var_name

    def __str__(self):
        return "W(" + self.memory_region + "-" + str(self.unique_id) + ", --> " + self.var_name + ")"

    def __hash__(self):
        return self.unique_id

    def toDict(self):
        result_dict = {}
        result_dict["memory_region"] = self.memory_region
        result_dict["unique_id"] = self.unique_id
        result_dict["var_name"] = self.var_name
        return result_dict


def write_data_access_from_dict(values: Dict[str, Any]) -> WriteDataAccess:
    return WriteDataAccess(values["memory_region"], values["unique_id"], values["var_name"])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from copy import deepcopy
from typing import List

from ..PEGraphX import LineID, LoopNode, PEGraphX, Node, EdgeType
from ..utils import correlation_coefficient

total = 0
before: List[float] = []
after: List[float] = []


def run_before(pet: PEGraphX):
    return pet


def run_after(pet: PEGraphX):
    for node in pet.all_nodes(LoopNode):
        check_pipeline(pet, node)

    print(f"Total: {total}")
    print(" ".join([str(x) for x in before]))
    print(" ".join([str(x) for x in after]))
    return pet


def check_pipeline(pet: PEGraphX, root: Node):
    """Tries to optimize dependencies for pipeline detection
    1. Deletes independent lines, that do not contribute to the pipeline
    2. Deletes similar CU (that have same dependencies), as those can be one step in the pipeline

    :param pet: PET graph
    :param root: current node
    :return: Pipeline scalar value
    """
    global total
    global before
    global after

    children_start_lines = [v.start_position() for v in pet.subtree_of_type(root, LoopNode)]

    loop_subnodes = [
        pet.node_at(t)
        for s, t, d in pet.out_edges(root.id, [EdgeType.CHILD, EdgeType.CALLSNODE])
        if is_pipeline_subnode(root, pet.node_at(t), children_start_lines)
    ]

    if len(loop_subnodes) < 3:
        return

    matrix = get_matrix(pet, root, loop_subnodes)
    initial_matrix = deepcopy(matrix)
    initial_coef = get_correlation_coefficient(matrix)

    if initial_coef < 0.999:
        total += 1
    independent_cus = get_independent_lines(matrix)
    delete_lines(matrix, loop_subnodes, independent_cus)

    no_indep_matrix = deepcopy(matrix)

    mergeable_cus = get_mergeable_nodes(matrix)
    delete_lines(matrix, loop_subnodes, mergeable_cus)

    new_coef = get_correlation_coefficient(matrix)
    if new_coef > initial_coef:
        before.append(initial_coef)
        after.append(new_coef)
        print("Pipeline improvement opportunity:")
        print("Node: " + root.id)
        print("Lines: " + root.start_position() + "-" + root.end_position())
        print("Independent lines:")
        independent_cus.sort()
        print(" ".join([str(x) for x in independent_cus]))
        print("Similar nodes:")
        mergeable_cus.sort()
        print(" ".join([str(x) for x in mergeable_cus]))
        print(f"matrix before: {initial_coef}")
        for i in range(0, len(initial_matrix)):
            print(" ".join([str(x) for x in initial_matrix[i]]))
        print(f"matrix after independent removed:")
        for i in range(0, len(no_indep_matrix)):
            print(" ".join([str(x) for x in no_indep_matrix[i]]))
        print(f"matrix after merged: {new_coef}")
        for i in range(0, len(matrix)):
            print(" ".join([str(x) for x in matrix[i]]))


def delete_lines(matrix, loop_nodes, lines):
    if lines:
        lines.sort(reverse=True)
        for i in range(0, len(lines)):
            del loop_nodes[lines[i]]
            del matrix[lines[i]]
            for j in range(0, len(matrix)):
                del matrix[j][lines[i]]


def get_independent_lines(matrix):
    res = []
    for i in range(0, len(matrix)):
        indep = True
        for j in range(0, len(matrix)):
            if i != j and (matrix[i][j] != 0 or matrix[j][i] != 0):
                indep = False
        if indep:
            res.append(i)
    return res


def get_mergeable_nodes(matrix):
    res = []
    for i in reversed(range(1, len(matrix))):
        if matrix[i] == matrix[i - 1]:
            same = True
            for j in range(1, len(matrix)):
                if matrix[j][i] != matrix[j][i - 1]:
                    same = False
            if same:
                res.append(i)
    return res


def get_matrix(pet: PEGraphX, root: Node, loop_subnodes: List[Node]) -> List[List[int]]:
    res: List[List[int]] = []
    for i in range(0, len(loop_subnodes)):
        res.append([])
        for j in range(0, len(loop_subnodes)):
            res[i].append(int(pet.depends_ignore_readonly(loop_subnodes[i], loop_subnodes[j], root)))
    return res


def get_correlation_coefficient(matrix):
    graph_vector = []
    for i in range(0, len(matrix) - 1):
        graph_vector.append(matrix[i + 1][i])

    pipeline_vector = []
    for i in range(0, len(matrix) - 1):
        pipeline_vector.append(1)

    min_weight = 1
    for i in range(0, len(matrix) - 1):
        for j in range(i + 1, len(matrix)):
            if matrix[i][j] == 1:
                node_weight = 1 - (j - i) / (len(matrix) - 1)
                if min_weight > node_weight > 0:
                    min_weight = node_weight

    if min_weight == 1:
        graph_vector.append(0)
        pipeline_vector.append(0)
    else:
        graph_vector.append(1)
        pipeline_vector.append(min_weight)
    return round(correlation_coefficient(graph_vector, pipeline_vector), 2)


def is_pipeline_subnode(root: Node, current: Node, children_start_lines: List[LineID]) -> bool:
    """Checks if node is a valid subnode for pipeline

    :param root: root node
    :param current: current node
    :param children_start_lines: start lines of children loops
    :return: true if valid
    """
    r_start = root.start_position()
    r_end = root.end_position()
    c_start = current.start_position()
    c_end = current.end_position()
    return not (
        c_start == r_start
        and c_end == r_start
        or c_start == r_end
        and c_end == r_end
        or c_start == c_end
        and c_start in children_start_lines
    )



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.suggestions.importers.do_all import (
    import_suggestion as import_doall,
)
from discopop_library.discopop_optimizer.suggestions.importers.reduction import (
    import_suggestion as import_reduction,
)


def import_suggestions(experiment: Experiment) -> nx.DiGraph:
    """Imports the suggestions specified in res into the graph stored in the given experiment and returns the modified graph"""

    # import do-all
    for do_all_suggestion in experiment.detection_result.patterns.do_all:
        experiment.optimization_graph = import_doall(
            experiment.optimization_graph, do_all_suggestion, experiment.get_next_free_node_id, experiment
        )

    # import reduction
    for reduction_suggestion in experiment.detection_result.patterns.reduction:
        experiment.optimization_graph = import_reduction(
            experiment.optimization_graph, reduction_suggestion, experiment.get_next_free_node_id, experiment
        )
    return experiment.optimization_graph



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
from typing import List, Tuple

from DEPRECATED_discopop_wizard.classes.Arguments import Arguments
from DEPRECATED_discopop_wizard.classes.ExecutionConfiguration import ExecutionConfiguration
from DEPRECATED_discopop_wizard.classes.Settings import load_from_config_file, Settings
from DEPRECATED_discopop_wizard.screens.execution import ExecutionView
from DEPRECATED_discopop_wizard.wizard import DiscoPoPConfigurationWizard


def execute_all_stored_configurations(args: Arguments, source_dir: str):
    wizard, settings, execution_configs = __load_data(args, source_dir)
    # execute all configurations
    execution_views: List[ExecutionView] = []
    for config in execution_configs:
        execution_views.append(ExecutionView(config, wizard, None, headless_mode=True))


def execute_tag_filtered_configurations(args: Arguments, source_dir: str):
    wizard, settings, execution_configs = __load_data(args, source_dir)
    # filter configurations by tags
    filtered_execution_configs: List[ExecutionConfiguration] = []
    for config in execution_configs:
        # get tags from config
        # get tags from arguments
        # if an overlap exists, the configurations shall be executed
        overlapping_tags = [tag for tag in config.get_tags() if tag in args.execute_configurations_with_tag]
        if len(overlapping_tags) > 0:
            filtered_execution_configs.append(config)

    # execute the filtered configurations
    execution_views: List[ExecutionView] = []
    for config in filtered_execution_configs:
        execution_views.append(ExecutionView(config, wizard, None, headless_mode=True))


def __load_data(
    args: Arguments, source_dir: str
) -> Tuple[DiscoPoPConfigurationWizard, Settings, List[ExecutionConfiguration]]:
    """Loads and returns a DiscoPoPConfigurationWizard, Settings, and a list of ExecutionConfigurations."""
    config_dir = os.path.join(source_dir, ".config")
    # load settings
    settings = load_from_config_file(config_dir)
    # create DiscoPoPWizard
    wizard = DiscoPoPConfigurationWizard(config_dir, args, headless_mode=True)

    # load execution configurations
    execution_configs: List[ExecutionConfiguration] = []
    for filename in os.listdir(os.path.join(config_dir, "execution_configurations")):
        if not filename.endswith(".json"):
            continue
        with open(os.path.join(os.path.join(config_dir, "execution_configurations"), filename), "r") as json_file:
            config = ExecutionConfiguration(wizard)
            config.init_from_json(json_file)
            execution_configs.append(config)

    return wizard, settings, execution_configs



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Optional, Set, List, cast

from sympy import Integer, Expr  # type: ignore

from discopop_explorer.PEGraphX import NodeID, PEGraphX, EdgeType
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.nodes.GenericNode import GenericNode
from discopop_library.discopop_optimizer.classes.types.DataAccessType import (
    WriteDataAccess,
    ReadDataAccess,
)


class Workload(GenericNode):
    """This class represents a generic node in the Optimization Graph"""

    sequential_workload: Optional[Expr]
    parallelizable_workload: Optional[Expr]
    cost_multiplier: CostModel

    def __init__(
        self,
        node_id: int,
        experiment,
        cu_id: Optional[NodeID],
        sequential_workload: Optional[Expr],
        parallelizable_workload: Optional[Expr],
        written_memory_regions: Optional[Set[WriteDataAccess]] = None,
        read_memory_regions: Optional[Set[ReadDataAccess]] = None,
    ):
        super().__init__(
            node_id,
            experiment,
            cu_id,
            written_memory_regions=written_memory_regions,
            read_memory_regions=read_memory_regions,
        )
        self.sequential_workload = sequential_workload
        self.parallelizable_workload = parallelizable_workload
        self.performance_model = CostModel(
            Integer(0) if self.parallelizable_workload is None else self.parallelizable_workload,
            Integer(0) if self.sequential_workload is None else self.sequential_workload,
        )
        self.cost_multiplier = CostModel(Integer(1), Integer(1))
        self.overhead = CostModel(Integer(0), Integer(0))

    def get_plot_label(self) -> str:
        if self.sequential_workload is not None:
            # return str(self.workload)
            return str(self.node_id) + "\n" + str(self.original_cu_id)
        else:
            return "WL"

    def get_hover_text(self) -> str:
        return (
            "WL: " + str(self.sequential_workload) + "\n"
            "Read: " + str([str(e) for e in self.read_memory_regions]) + "\n"
            "Write: " + str([str(e) for e in self.written_memory_regions]) + "\n"
            "Branch: " + str(self.branch_affiliation) + "\n"
            "Device ID: " + str(self.device_id)
        )

    def get_cost_model(self, experiment, all_function_nodes, current_device) -> CostModel:
        """Performance model of a workload consists of the workload itself + the workload of called functions.
        Individual Workloads are assumed to be not parallelizable.
        Workloads of called functions are added as encountered.
        Workloads of Loop etc. are parallelizable."""

        cm: CostModel

        if self.sequential_workload is None:
            cm = (
                CostModel(Integer(1), Integer(0))
                .parallelizable_multiply_combine(self.cost_multiplier)
                .parallelizable_plus_combine(self.overhead)
                .parallelizable_plus_combine(
                    self.__get_costs_of_function_call(experiment, all_function_nodes, current_device)
                )
            )
        else:
            cm = (
                CostModel(
                    current_device.get_estimated_execution_time_in_micro_seconds(self.parallelizable_workload, True),
                    current_device.get_estimated_execution_time_in_micro_seconds(self.sequential_workload, True),
                )
                .parallelizable_multiply_combine(self.cost_multiplier)
                .parallelizable_plus_combine(self.overhead)
                .parallelizable_plus_combine(
                    self.__get_costs_of_function_call(experiment, all_function_nodes, current_device)
                )
            )

        # substitute Expr(0) with Integer(0)
        cm.parallelizable_costs = cm.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})
        cm.sequential_costs = cm.sequential_costs.subs({Expr(Integer(0)): Integer(0)})

        return cm

    def __get_costs_of_function_call(self, experiment, all_function_nodes, current_device) -> CostModel:
        """Check if the node performs a function call and returns the total costs for these."""
        total_costs = CostModel(Integer(0), Integer(0))
        # get CUIDs of called functions
        if self.original_cu_id is not None:
            called_cu_ids: List[str] = [
                str(t)
                for s, t, d in cast(PEGraphX, experiment.detection_result.pet).out_edges(
                    self.original_cu_id, EdgeType.CALLSNODE
                )
            ]
            # filter for called FunctionRoots
            called_function_nodes = [fr for fr in all_function_nodes if str(fr.original_cu_id) in called_cu_ids]
            # remove duplicates
            called_function_nodes = list(set(called_function_nodes))
            # add costs of called function nodes to total costs
            for called_function_root in called_function_nodes:
                total_costs = total_costs.parallelizable_plus_combine(
                    called_function_root.get_cost_model(experiment, all_function_nodes, current_device)
                )

        return total_costs

    def register_child(self, other, experiment, all_function_nodes, current_device):
        """Registers a child node for the given model.
        Does not modify the stored model in self or other."""
        # since workloads do not modify their children, the performance model of other is simply added to self.
        return self.performance_model.parallelizable_plus_combine(other)

    def register_successor(self, other):
        """Registers a successor node for the given model.
        Does not modify the stored model in self or other."""
        # sequential composition is depicted by simply adding the performance models
        return self.performance_model.parallelizable_plus_combine(other)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, List, Tuple, Union

import numpy as np
from scipy.interpolate import LinearNDInterpolator  # type: ignore

from .Microbench import (
    Microbench,
    MicrobenchType,
    MicrobenchDimension,
    MicrobenchCoordinate,
)
from .PureDataMicrobench import PureDataMicrobench


# This class uses Delaunay Interpolation to create a microbench model from measurements. No extrapolation is possible
class DelaunayInterpolatedMicrobench(Microbench):
    data: PureDataMicrobench

    __isInterpolated = False

    def __init__(
        self,
        data: Union[
            PureDataMicrobench,
            Dict[
                MicrobenchType,
                Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],
            ],
        ],
        removeOutliers: bool = False,
    ):
        if isinstance(data, PureDataMicrobench):
            self.data = data
        else:
            self.data = PureDataMicrobench(data)

        if removeOutliers:
            self.data.removeOutliers()

    def removeZeroParameters(self):
        self.data.removeZeroParameters()
        self.__isInterpolated = False

    def clampValues(self, min: float = 0.0, max: float = float("inf")):
        self.data.clampValues()
        self.__isInterpolated = False

    def removeOutliers(self):
        self.data.removeOutliers()
        self.__isInterpolated = False

    def useMedian(self):
        self.data.useMedian()
        self.__isInterpolated = False

    def useMean(self):
        self.data.useMean()
        self.__isInterpolated = False

    def __getTuples(self, benchType: MicrobenchType, benchDim: MicrobenchDimension):
        tuples: List[Tuple[int, Union[int, float], int, float]] = []
        for benchCoord, values in self.data[benchType][benchDim].items():
            tuples.append((*benchCoord, np.median(values).item()))
        return tuples

    def __interpolate(self):
        def __createInterpolator(tuples):
            coords = []
            values = []
            for t in tuples:
                coords.append(t[:3])
                values.append(t[3])
            return LinearNDInterpolator(coords, values, fill_value=float("nan"))

        self.interpolator = {
            type: {dim: __createInterpolator(self.__getTuples(type, dim)) for dim in dimMap.keys()}
            for type, dimMap in self.data.getMeasurements().items()
        }

    def getMeasurements(self):
        return self.data.getMeasurements()

    def toJSON(self):
        return self.data.toJSON()

    def evaluateInterpolation(
        self,
        benchType: MicrobenchType,
        benchDim: MicrobenchDimension,
        benchCoord: Union[MicrobenchCoordinate, Tuple[int, float, float]],
    ):
        if not self.__isInterpolated:
            self.__interpolate()
            self.__isInterpolated = True
        return self.interpolator[benchType][benchDim](benchCoord) / 1000000.0  # convert microseconds to seconds



import os
import pathlib
import unittest

import jsonpickle

from discopop_library.result_classes.DetectionResult import DetectionResult
from test.utils.validator_classes.DoAllInfoForValidation import DoAllInfoForValidation


class TestMethods(unittest.TestCase):
    def test(self):
        current_dir = pathlib.Path(__file__).parent.resolve()
        discopop_dir = os.path.join(current_dir, "..", "..", "..", "..", "..")
        build_dir = os.path.join(discopop_dir, "build")

        src_dir = os.path.join(current_dir, "src")
        os.chdir(src_dir)

        # create FileMapping
        os.system("" + os.path.join(build_dir, "scripts", "dp-fmap"))

        # build
        # make_command = "DP_FM_PATH=" + os.path.join(src_dir, "FileMapping.txt") + " "
        make_command = ""
        make_command += "CC=" + os.path.join(build_dir, "scripts", "CC_wrapper.sh") + " "
        make_command += "CXX=" + os.path.join(build_dir, "scripts", "CXX_wrapper.sh") + " "
        # make_command += "LINKER=" + os.path.join(build_dir, "scripts", "LINKER_wrapper.sh") + " "
        make_command += "make "
        os.system(make_command)
        # execute instrumented program
        os.system("./prog")
        # execute DiscoPoP analysis
        os.chdir(".discopop")
        os.system("discopop_explorer")
        os.system("discopop_optimizer -x")
        os.chdir("..")
        # validate results
        self.validate_results(current_dir, src_dir)
        # clean environment
        os.system("make veryclean")

    def validate_results(self, test_dir, src_dir):
        """Check that not collapse has been identified"""
        
        # load test output
        test_output_file = os.path.join(src_dir, ".discopop", "optimizer", "detection_result_dump.json")
        with open(test_output_file, "r") as f:
            tmp_str = f.read()
        test_output: DetectionResult = jsonpickle.decode(tmp_str)

        # check identified DoAllInfo objects for collapse clauses > 1
        for do_all_info in test_output.patterns.do_all:
            self.assertTrue(do_all_info.collapse_level <= 1)





# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List

from discopop_explorer.pattern_detectors.PatternBase import PatternBase
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.do_all_detector import DoAllInfo
from discopop_explorer.pattern_detectors.geometric_decomposition_detector import GDInfo
from discopop_explorer.pattern_detectors.pipeline_detector import PipelineInfo
from discopop_explorer.pattern_detectors.reduction_detector import ReductionInfo


class PatternStorage(object):
    reduction: List[ReductionInfo]
    do_all: List[DoAllInfo]
    pipeline: List[PipelineInfo]
    geometric_decomposition: List[GDInfo]
    task: List[PatternInfo]
    simple_gpu: List[PatternInfo]
    combined_gpu: List[PatternInfo]
    optimizer_output: List[PatternBase]

    def __init__(self):
        self.optimizer_output = []



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
import random
from typing import List, Dict, cast, Set, Optional

import networkx as nx  # type: ignore
import sympy  # type: ignore
from sympy import Integer, Expr, Symbol  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_in_mutex_edges,
    get_out_mutex_edges,
    get_successors,
    get_children,
    get_edge_data,
    get_requirements,
    get_out_options,
    get_in_options,
    get_all_parents,
    get_all_function_nodes,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


def get_performance_models_for_functions(
    experiment: Experiment, graph: nx.DiGraph, restrict_to_decisions: Optional[Set[int]] = None
) -> Dict[FunctionRoot, List[CostModel]]:
    performance_models: Dict[FunctionRoot, List[CostModel]] = dict()
    # get called FunctionRoots from cu ids
    all_function_nodes = [
        cast(FunctionRoot, data_at(experiment.optimization_graph, fn_id))
        for fn_id in get_all_function_nodes(experiment.optimization_graph)
    ]

    for node_id in graph.nodes:
        node_data = graph.nodes[node_id]["data"]
        node_data.node_id = node_id  # fix potential mismatches due to node copying

        if isinstance(node_data, FunctionRoot):
            # start the collection at the first child of the function
            for child_id in get_children(graph, node_id):
                performance_models[node_data] = get_node_performance_models(
                    experiment, graph, child_id, set(), all_function_nodes, restrict_to_decisions=restrict_to_decisions
                )

            # At this point, decisions are restricted to the specified parallelization or the sequential version.
            # Restrict them to the exact case specified in restrict_to_decisions
            if restrict_to_decisions is not None:
                to_be_removed: List[int] = []
                for idx, cost_model in enumerate(performance_models[node_data]):
                    for decision in cost_model.path_decisions:
                        if decision not in restrict_to_decisions:
                            to_be_removed.append(idx)
                            break
                for idx in sorted(to_be_removed, reverse=True):
                    del performance_models[node_data][idx]

            # filter out NaN - Models
            performance_models[node_data] = [
                model for model in performance_models[node_data] if model.parallelizable_costs != sympy.nan
            ]

    return performance_models


def get_node_performance_models(
    experiment: Experiment,
    graph: nx.DiGraph,
    node_id: int,
    visited_nodes: Set[int],
    all_function_nodes: List[FunctionRoot],
    restrict_to_decisions: Optional[Set[int]] = None,
    do_not_allow_decisions: Optional[Set[int]] = None,
    get_single_random_model: bool = False,
    ignore_node_costs: Optional[List[int]] = None,
    current_device_id=None,
) -> List[CostModel]:
    """Returns the performance models for the given node.
    If a set of decision is specified for restrict_to_decisions, only those non-sequential decisions will be allowed.
    Caution: List might be empty!
    """
    result_list: List[CostModel] = []
    successors = get_successors(graph, node_id)
    successor_count = len(successors)
    node_data = data_at(graph, node_id)
    if node_data.execute_in_parallel:
        current_device_id = node_data.device_id
    visited_nodes.add(node_id)

    # consider performance models of children
    children_models = get_performance_models_for_children(
        experiment,
        graph,
        node_id,
        copy.deepcopy(visited_nodes),
        all_function_nodes,
        restrict_to_decisions=restrict_to_decisions,
        do_not_allow_decisions=do_not_allow_decisions,
        get_single_random_model=get_single_random_model,
    )

    if len(children_models) == 0:
        if ignore_node_costs is not None:
            if node_data.node_id in ignore_node_costs:
                children_models = [CostModel(Integer(0), Integer(0))]
        else:
            children_models = [
                node_data.get_cost_model(
                    experiment,
                    all_function_nodes,
                    experiment.get_system().get_device(current_device_id),
                )
            ]

    else:
        if ignore_node_costs is not None:
            if node_data.node_id in ignore_node_costs:
                tmp_node_cost_model = CostModel(Integer(0), Integer(0))
        else:
            tmp_node_cost_model = node_data.get_cost_model(
                experiment,
                all_function_nodes,
                experiment.get_system().get_device(current_device_id),
            )

        for idx, child_model in enumerate(children_models):
            if ignore_node_costs is not None:
                if node_data.node_id not in ignore_node_costs:
                    children_models[idx] = tmp_node_cost_model.register_child(
                        child_model,
                        node_data,
                        experiment,
                        all_function_nodes,
                        experiment.get_system().get_device(current_device_id),
                    )
            else:
                children_models[idx] = tmp_node_cost_model.register_child(
                    child_model,
                    node_data,
                    experiment,
                    all_function_nodes,
                    experiment.get_system().get_device(current_device_id),
                )

    # construct the performance models
    if successor_count >= 1:
        removed_successors = False
        if get_single_random_model and successor_count > 1:
            # pick only a single successor
            successors = [random.choice(successors)]
            removed_successors = True

        for children_model in children_models:
            for successor in successors:
                # ## CHECK REQUIREMENTS ##
                # check if successor validates a requirements edge to restrain the created combinations
                # 1.1. check if optionEdge between any node in visited_nodes and successor exists
                # 1.2. if so, check if option edge to other node in visited nodes exists
                # 1.3. if so, check if a requirements edge between both option exists.
                # 1.4. if not, the path is not valid since two options for the same
                #      source code location would be selected
                path_invalid = False
                # 1.1
                #               if successor in [109, 110, 111, 112, 113, 114]:
                #                      print("VISITED NODES: ", visited_nodes)
                for visited_node_id in visited_nodes:
                    options = get_out_mutex_edges(graph, visited_node_id)
                    if successor in options:
                        # 1.2
                        visited_options = [opt for opt in options if opt in visited_nodes]
                        if len(visited_options) > 0:
                            # 1.3
                            for vo in visited_options:
                                # 1.4
                                if successor not in get_requirements(graph, vo):
                                    path_invalid = True
                                    break
                    if path_invalid:
                        break
                if path_invalid:
                    continue

                # 2 check if a sibling of successor exists which has a requirements edge to a visited node
                # 2.1 check if an incoming or outgoing option edge exists, get the node id for the sequential version
                # 2.2 for all parallelization options
                # 2.3 check if a requirements edge to a visited node exists
                # 2.4 if so, stop if successor is NOT the parallelization option with the requirements edge
                # 2.1
                for sibling in successors:
                    sequential_version_ids = []
                    if len(get_out_mutex_edges(graph, sibling)) > 0:
                        sequential_version_ids = [sibling]
                    else:
                        for seq in get_in_mutex_edges(graph, sibling):
                            sequential_version_ids.append(seq)
                    # 2.2
                    for seq in sequential_version_ids:
                        for option in get_out_mutex_edges(graph, seq):
                            if option == successor:
                                continue
                            # 2.3
                            for visited_req in [req for req in get_requirements(graph, option) if req in visited_nodes]:
                                # 2.4
                                if visited_req != successor:
                                    path_invalid = True
                                    break
                            if path_invalid:
                                break
                    if path_invalid:
                        break

                # do not allow nested parallelization suggestions on devices of type GPU
                if True:  # option to disable this check
                    combined_visited_nodes = copy.deepcopy(visited_nodes)
                    combined_visited_nodes.add(successor)
                    gpu_suggestions = [
                        node_id
                        for node_id in combined_visited_nodes
                        if isinstance(
                            experiment.get_system().get_device(data_at(graph, node_id).device_id),
                            GPU,
                        )
                    ]
                    # check if two suggestions are in a contained-in relation
                    for suggestion_1 in gpu_suggestions:
                        all_parents = get_all_parents(graph, suggestion_1)
                        for suggestion_2 in gpu_suggestions:
                            if suggestion_1 == suggestion_2:
                                continue
                            if suggestion_2 in all_parents:
                                path_invalid = True
                                break
                        if path_invalid:
                            break

                # check if the current decision invalidates decision requirements, if some are specified
                if restrict_to_decisions is not None:
                    if not (successor in restrict_to_decisions or data_at(graph, successor).suggestion is None):
                        path_invalid = True
                    if not path_invalid:
                        if data_at(graph, successor).suggestion is None:
                            # if the sequential "fallback" has been used, check if a different option is specifically
                            # mentioned in restrict_to_decisions. If so, the sequential fallback shall be ignored.
                            options = get_out_mutex_edges(graph, successor)
                            restricted_options = [opt for opt in options if opt in restrict_to_decisions]
                            if len(restricted_options) != 0:
                                # do not use he sequential fallback since a required option exists
                                path_invalid = True

                if do_not_allow_decisions is not None:
                    if successor in do_not_allow_decisions:
                        path_invalid = True

                if path_invalid:
                    continue

                # ## END OF REQUIREMENTS CHECK ##

                combined_model = children_model
                # add transfer costs
                transfer_costs_model = get_edge_data(graph, node_id, successor).get_cost_model()
                combined_model = combined_model.parallelizable_plus_combine(transfer_costs_model)

                # if the successor is "determined" by a path decision, add path decision to the combined model
                if len(successors) > 1 or removed_successors:
                    combined_model.path_decisions.append(successor)
                # append the model of the successor
                for model in get_node_performance_models(
                    experiment,
                    graph,
                    successor,
                    copy.deepcopy(visited_nodes),
                    all_function_nodes,
                    restrict_to_decisions=restrict_to_decisions,
                    do_not_allow_decisions=do_not_allow_decisions,
                    get_single_random_model=get_single_random_model,
                    ignore_node_costs=ignore_node_costs,
                ):
                    tmp = combined_model.parallelizable_plus_combine(model)
                    tmp.path_decisions += [d for d in model.path_decisions if d not in tmp.path_decisions]
                    result_list.append(tmp)
        if len(result_list) >= 1:
            return result_list

    # successor count == 0 or successor count > 1
    return children_models


def get_performance_models_for_children(
    experiment: Experiment,
    graph: nx.DiGraph,
    node_id: int,
    visited_nodes: Set[int],
    all_function_nodes: List[FunctionRoot],
    restrict_to_decisions: Optional[Set[int]] = None,
    do_not_allow_decisions: Optional[Set[int]] = None,
    get_single_random_model: bool = False,
) -> List[CostModel]:
    """Construct a performance model for the children of the given node, or return None if no children exist"""
    # todo: consider children
    child_models: List[CostModel] = []

    # create all combinations for models from children
    first_iteration = True
    for child_id in get_children(graph, node_id):
        if first_iteration:
            first_iteration = False
            for model in get_node_performance_models(
                experiment,
                graph,
                child_id,
                copy.deepcopy(visited_nodes),
                all_function_nodes,
                restrict_to_decisions=restrict_to_decisions,
                do_not_allow_decisions=do_not_allow_decisions,
                get_single_random_model=get_single_random_model,
            ):
                # initialize list of child models
                child_models.append(model)
        else:
            # create "product set" of child models
            product_set = []
            for model in get_node_performance_models(
                experiment,
                graph,
                child_id,
                copy.deepcopy(visited_nodes),
                all_function_nodes,
                restrict_to_decisions=restrict_to_decisions,
                do_not_allow_decisions=do_not_allow_decisions,
                get_single_random_model=get_single_random_model,
            ):
                temp_models = [cm.parallelizable_plus_combine(model) for cm in child_models]
                product_set += temp_models
            child_models = product_set
    return child_models


def print_introduced_symbols_per_node(graph: nx.DiGraph):
    print("Introduced Symbols:")
    for node_id in graph.nodes:
        print("NodeID: ", node_id)
        for symbol in data_at(graph, node_id).introduced_symbols:
            print("\t: ", symbol)
    print()


def get_random_path(
    experiment: Experiment, graph: nx.DiGraph, root_id: int, must_contain: Optional[Set[int]] = None
) -> CostModel:
    # get called FunctionRoots from cu ids
    all_function_nodes = [
        cast(FunctionRoot, data_at(experiment.optimization_graph, fn_id))
        for fn_id in get_all_function_nodes(experiment.optimization_graph)
    ]
    random_models = get_node_performance_models(
        experiment,
        graph,
        root_id,
        set(),
        all_function_nodes,
        restrict_to_decisions=must_contain,
        get_single_random_model=True,
        ignore_node_costs=[cast(FunctionRoot, data_at(graph, root_id)).node_id],
    )
    # filter out NaN - Models
    random_models = [model for model in random_models if model.parallelizable_costs != sympy.nan]
    return random.choice(random_models)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import cast
from discopop_library.ConfigProvider.ConfigProviderArguments import ConfigProviderArguments
from discopop_library.ConfigProvider.assets.build_config import DP_BUILD, DP_SOURCE, LLVM_BIN_DIR  # type: ignore


def run(arguments: ConfigProviderArguments) -> str:
    """Returns the contents of the written build_config.txt"""

    if arguments.return_dp_build_dir:
        return cast(str, DP_BUILD)  # type: ignore
    elif arguments.return_dp_source_dir:
        return cast(str, DP_SOURCE)  # type: ignore
    elif arguments.return_llvm_bin_dir:
        return cast(str, LLVM_BIN_DIR)  # type: ignore
    else:
        raise ValueError("No valid operation for execution configuration: \n" + str(arguments))



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os
from pathlib import Path
from typing import Dict, Tuple, Set, Optional, List, Any

import networkx as nx  # type: ignore
from sympy import Integer, Symbol, Expr, Float  # type: ignore

from discopop_explorer.PEGraphX import MemoryRegion
from discopop_library.MemoryRegions.utils import get_sizes_of_memory_regions
from discopop_library.PathManagement.PathManagement import load_file_mapping
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.system.System import System
from discopop_library.result_classes.DetectionResult import DetectionResult


class Experiment(object):
    ## SETTINGS
    # todo: convert Costs into estimated runtime, removes need for high overhead weight
    workload_overhead_weight = Integer(1500)
    do_all_overhead_weight_by_device: Dict[int, Expr] = {
        0: Integer(300),
        1: Integer(300),
        2: Integer(300),
    }
    reduction_overhead_weight_by_device: Dict[int, Expr] = {
        0: Integer(300),
        1: Integer(300),
        2: Integer(300),
    }

    ## END OF SETTINGS

    __system: System

    # all free symbols will be added to this list for simple retrieval and user query
    free_symbols: Set[Symbol] = set()
    substitutions: Dict[Symbol, Expr] = dict()
    sorted_free_symbols: List[Symbol] = []
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]] = dict()
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution] = dict()
    # value suggestions for all free symbols will be stored in this dictionary
    suggested_values: Dict[Symbol, Expr] = dict()

    __memory_region_sizes: Dict[MemoryRegion, int]  # sizes in Bytes

    project_path: Path
    discopop_output_path: Path
    profiler_dir: Path
    code_export_path: Path

    file_mapping: Dict[int, Path]  # file-mapping

    arguments: OptimizerArguments

    detection_result: DetectionResult

    function_models: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject, str]]]
    selected_paths_per_function: Dict[FunctionRoot, Tuple[CostModel, ContextObject]]

    optimization_graph: nx.DiGraph
    next_free_node_id: int
    suggestion_to_node_ids_dict: Dict[int, List[int]]
    node_id_to_suggestion_dict: Dict[int, int]

    def __init__(
        self,
        file_mapping: Dict[int, Path],
        system: System,
        detection_result: DetectionResult,
        profiler_dir: str,
        arguments: OptimizerArguments,
    ):
        self.__system = system
        self.detection_result = detection_result
        self.arguments = arguments

        self.__memory_region_sizes = get_sizes_of_memory_regions(
            set(),
            os.path.join(profiler_dir, "memory_regions.txt"),
            return_all_memory_regions=True,
        )

        self.file_mapping = file_mapping
        self.function_models = dict()
        self.selected_paths_per_function = dict()
        self.suggestion_to_node_ids_dict = dict()
        self.node_id_to_suggestion_dict = dict()

        # collect free symbols from system
        for free_symbol, value_suggestion in system.get_free_symbols():
            self.register_free_symbol(free_symbol, value_suggestion)

    def get_memory_region_size(
        self, memory_region: MemoryRegion, use_symbolic_value: bool = False
    ) -> Tuple[Expr, Expr]:
        if memory_region not in self.__memory_region_sizes:
            self.__memory_region_sizes[memory_region] = 8  # assume 8 Bytes for unknown sizes

        if use_symbolic_value:
            symbolic_memory_region_size = Symbol("mem_reg_size_" + str(memory_region))
            # register the symbolic value in the environment
            self.register_free_symbol(
                symbolic_memory_region_size,
                value_suggestion=Integer(self.__memory_region_sizes[memory_region]),
            )
            return symbolic_memory_region_size, Integer(self.__memory_region_sizes[memory_region])
        else:
            return Integer(self.__memory_region_sizes[memory_region]), Integer(0)

    def register_free_symbol(self, symbol: Symbol, value_suggestion: Optional[Expr] = None):
        self.free_symbols.add(symbol)
        if value_suggestion is not None:
            self.suggested_values[symbol] = value_suggestion

    def get_next_free_node_id(self):
        buffer = self.next_free_node_id
        self.next_free_node_id += 1
        return buffer

    def get_system(self) -> System:
        return self.__system



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
import itertools
from multiprocessing import Pool
from typing import Any, ClassVar, Dict, List, Optional, cast, Set, Tuple

import matplotlib  # type: ignore
import matplotlib.pyplot as plt  # type:ignore
import networkx as nx  # type: ignore
import tqdm  # type: ignore

from discopop_explorer.PEGraphX import MemoryRegion, NodeID
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.classes.edges.DataFlowEdge import DataFlowEdge
from discopop_library.discopop_optimizer.classes.edges.MutuallyExclusiveEdge import MutuallyExclusiveEdge
from discopop_library.discopop_optimizer.classes.edges.ChildEdge import ChildEdge
from discopop_library.discopop_optimizer.classes.edges.GenericEdge import GenericEdge
from discopop_library.discopop_optimizer.classes.edges.OptionEdge import OptionEdge
from discopop_library.discopop_optimizer.classes.edges.RequirementEdge import RequirementEdge
from discopop_library.discopop_optimizer.classes.edges.SuccessorEdge import SuccessorEdge
from discopop_library.discopop_optimizer.classes.edges.TemporaryEdge import TemporaryEdge
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.nodes.GenericNode import GenericNode
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at

from networkx.drawing.nx_pydot import graphviz_layout  # type: ignore


def get_nodes_from_cu_id(graph: nx.DiGraph, cu_node_id: NodeID) -> List[int]:
    result_list: List[int] = []
    for node_id in graph.nodes:
        if cast(GenericNode, graph.nodes[node_id]["data"]).original_cu_id == cu_node_id:
            result_list.append(node_id)
    return result_list


def get_edge_data(graph: nx.DiGraph, source: int, target: int) -> GenericEdge:
    return cast(GenericEdge, graph.edges[(source, target)]["data"])


def get_successors(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the successors of the given node"""
    return [edge[1] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], SuccessorEdge)]


def get_temporary_successors(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the temporary successors of the given node"""
    return [edge[1] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], TemporaryEdge)]


def get_predecessors(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the predecessors of the given node"""
    return [edge[0] for edge in graph.in_edges(node_id, data="data") if isinstance(edge[2], SuccessorEdge)]


def get_children(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the children of the given node"""
    return [edge[1] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], ChildEdge)]


def get_in_child_edges_sources(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the children of the given node"""
    return [edge[0] for edge in graph.in_edges(node_id, data="data") if isinstance(edge[2], ChildEdge)]


def get_out_options(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the parallelization options of the given node"""
    return [edge[1] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], OptionEdge)]


def get_in_options(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the parallelization options of the given node"""
    return [edge[0] for edge in graph.in_edges(node_id, data="data") if isinstance(edge[2], OptionEdge)]


def get_in_mutex_edges(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids which are mutually exclusive to the current node_id"""
    return [edge[0] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], MutuallyExclusiveEdge)]


def get_out_mutex_edges(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids which are mutually exclusive to the current node_id"""
    return [edge[1] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], MutuallyExclusiveEdge)]


def get_requirements(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns a list of node ids for the requirements of the parallelization option in the given node"""
    return [edge[1] for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], RequirementEdge)]


def has_temporary_successor(graph: nx.DiGraph, node_id: int) -> bool:
    """Checks whether the given node has outgoing temporary successor edges"""
    return len([edge for edge in graph.out_edges(node_id, data="data") if isinstance(edge[2], TemporaryEdge)]) > 0


def show(graph, show_dataflow: bool = True, show_mutex_edges: bool = True):
    """Plots the graph

    :return:
    """
    matplotlib.use("TkAgg")
    fig, ax = plt.subplots()
    try:
        # pos = nx.planar_layout(graph)  # good
        pos = graphviz_layout(graph, prog="dot")
    except nx.exception.NetworkXException:
        try:
            # fallback layouts
            pos = nx.shell_layout(graph)  # maybe
        except nx.exception.NetworkXException:
            pos = nx.random_layout(graph)

    drawn_nodes = set()
    nodes_lists: Dict[Any, Any] = dict()
    node_ids: Dict[Any, Any] = dict()
    node_insertion_sequence: List[Any] = []
    # draw nodes
    node_insertion_sequence.append(FunctionRoot)
    nodes_lists[FunctionRoot] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="#ff5151",
        node_shape="d",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), FunctionRoot)],
    )
    node_ids[FunctionRoot] = [n for n in graph.nodes if isinstance(data_at(graph, n), FunctionRoot)]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), FunctionRoot)])

    node_insertion_sequence.append(Loop)
    nodes_lists[Loop] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="#ff5151",
        node_shape="s",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), Loop)],
    )
    node_ids[Loop] = [n for n in graph.nodes if isinstance(data_at(graph, n), Loop)]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), Loop)])

    node_insertion_sequence.append(ContextNode)
    nodes_lists[ContextNode] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="yellow",
        node_shape="s",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), ContextNode)],
    )
    node_ids[ContextNode] = [n for n in graph.nodes if isinstance(data_at(graph, n), ContextNode)]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), ContextNode)])

    node_insertion_sequence.append(Workload)
    nodes_lists[Workload] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="#2B85FD",
        node_shape="o",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), Workload) and n not in drawn_nodes],
    )
    node_ids[Workload] = [n for n in graph.nodes if isinstance(data_at(graph, n), Workload) and n not in drawn_nodes]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), Workload) and n not in drawn_nodes])

    # id as label
    labels = {}
    for n in graph.nodes:
        labels[n] = graph.nodes[n]["data"].get_plot_label()
    nx.draw_networkx_labels(graph, pos, labels, font_size=7, ax=ax)

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="black",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], SuccessorEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="red",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], ChildEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="blue",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], TemporaryEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="pink",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], OptionEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="yellow",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], RequirementEdge)],
    )

    if show_mutex_edges:
        nx.draw_networkx_edges(
            graph,
            pos,
            ax=ax,
            edge_color="orange",
            edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], MutuallyExclusiveEdge)],
        )

    if show_dataflow:
        nx.draw_networkx_edges(
            graph,
            pos,
            ax=ax,
            edge_color="green",
            edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], DataFlowEdge)],
        )

    # define tool tip style when hovering
    # based on https://stackoverflow.com/questions/61604636/adding-tooltip-for-nodes-in-python-networkx-graph
    annot = ax.annotate(  # type: ignore
        "",
        xy=(0, 0),
        xytext=(20, 20),
        textcoords="offset points",
        bbox=dict(boxstyle="round", fc="w"),
        arrowprops=dict(arrowstyle="->"),
    )
    annot.set_visible(False)

    idx_to_node_dict = {}
    for idx, node in enumerate(graph.nodes):
        idx_to_node_dict[idx] = node

    def update_annot(ind, node_ids_list):
        node_idx = ind["ind"][0]
        node_id = node_ids_list[node_idx]
        xy = pos[node_id]
        annot.xy = xy
        node_attr = {"node": node_id}
        node_attr.update(graph.nodes[node_id])
        text = data_at(graph, node_id).get_hover_text()
        annot.set_text(text)

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            for node_type in node_insertion_sequence:
                nodes = nodes_lists[node_type]
                try:
                    cont, ind = nodes.contains(event)
                    if cont:
                        update_annot(ind, node_ids[node_type])
                        annot.set_visible(True)
                        fig.canvas.draw_idle()
                    else:
                        if vis:
                            annot.set_visible(False)
                            fig.canvas.draw_idle()
                except TypeError:
                    pass

    fig.canvas.mpl_connect("motion_notify_event", hover)  # type: ignore

    plt.show()


def add_successor_edge(graph: nx.DiGraph, source_id: int, target_id: int):
    edge_data = SuccessorEdge()
    graph.add_edge(source_id, target_id, data=edge_data)


def add_child_edge(graph: nx.DiGraph, source_id: int, target_id: int):
    edge_data = ChildEdge()
    graph.add_edge(source_id, target_id, data=edge_data)


def add_temporary_edge(graph: nx.DiGraph, source_id: int, target_id: int):
    edge_data = TemporaryEdge()
    graph.add_edge(source_id, target_id, data=edge_data)


def add_dataflow_edge(graph: nx.DiGraph, source_id: int, target_id: int):
    edge_data = DataFlowEdge()
    graph.add_edge(source_id, target_id, data=edge_data)


def delete_outgoing_temporary_edges(graph: nx.DiGraph, source_id: int):
    to_be_deleted = set()
    for edge in graph.out_edges(source_id):
        edge_data = graph.edges[edge]["data"]
        if isinstance(edge_data, TemporaryEdge):
            to_be_deleted.add(edge)
    for edge in to_be_deleted:
        remove_edge(graph, edge[0], edge[1])


def redirect_edge(
    graph: nx.DiGraph,
    old_source_id: int,
    new_source_id: int,
    old_target_id: int,
    new_target_id: int,
):
    edge_data = graph.edges[(old_source_id, old_target_id)]["data"]
    remove_edge(graph, old_source_id, old_target_id)
    graph.add_edge(new_source_id, new_target_id, data=edge_data)


def remove_edge(graph: nx.DiGraph, source_id: int, target_id: int):
    graph.remove_edge(source_id, target_id)


def convert_temporary_edges(graph: nx.DiGraph):
    """Convert temporary edges to Successor edges"""
    for edge in graph.edges:
        edge_data = graph.edges[edge]["data"]
        if isinstance(edge_data, TemporaryEdge):
            graph.edges[edge]["data"] = edge_data.convert_to_successor_edge()


def convert_temporary_edge(graph: nx.DiGraph, source_id: int, target_id: int):
    """Converts a single temporary edge to a successor edge"""
    edge_data = graph.edges[(source_id, target_id)]["data"]
    if isinstance(edge_data, TemporaryEdge):
        graph.edges[(source_id, target_id)]["data"] = edge_data.convert_to_successor_edge()


def get_all_function_nodes(graph: nx.DiGraph) -> List[int]:
    result_set: Set[int] = set()
    for node_id in graph.nodes:
        if type(graph.nodes[node_id]["data"]) == FunctionRoot:
            result_set.add(node_id)
    return list(result_set)


def get_all_nodes_in_function(graph: nx.DiGraph, function_id: int) -> List[int]:
    result_list: List[int] = []
    for node_id in graph.nodes:
        if function_id in get_all_parents(graph, node_id):
            result_list.append(node_id)
    return result_list


def get_nodes_by_functions(graph: nx.DiGraph) -> Dict[int, List[int]]:
    result_dict: Dict[int, List[int]] = dict()
    function_nodes = set(get_all_function_nodes(graph))
    for node in graph.nodes:
        parents = get_all_parents(graph, node)
        parent_functions = set(parents).intersection(function_nodes)
        for pf in parent_functions:
            if pf not in result_dict:
                result_dict[pf] = []
            result_dict[pf].append(node)
    return result_dict


def get_all_loop_nodes(graph: nx.DiGraph) -> List[int]:
    result_set: Set[int] = set()
    for node_id in graph.nodes:
        if type(graph.nodes[node_id]["data"]) == Loop:
            result_set.add(node_id)
    return list(result_set)


def get_read_and_written_data_from_subgraph(
    graph: nx.DiGraph, node_id: int, ignore_successors: bool = False
) -> Tuple[Set[MemoryRegion], Set[MemoryRegion]]:
    """Collect written and read memory regions for the node itself, it's successor and it's children. Used in the data flow calculation step."""
    read_memory_regions: Set[MemoryRegion] = set()
    written_memory_regions: Set[MemoryRegion] = set()
    # collect reads and writes from successors and children
    subgraph = get_children(graph, node_id)
    if not ignore_successors:
        subgraph += get_successors(graph, node_id)
    for successor in subgraph:
        reads, writes = get_read_and_written_data_from_subgraph(graph, successor)
        read_memory_regions.update(reads)
        written_memory_regions.update(writes)
    # add reads and writes of the node itself
    node_data = data_at(graph, node_id)
    read_memory_regions.update([read_access.memory_region for read_access in node_data.read_memory_regions])
    written_memory_regions.update([write_access.memory_region for write_access in node_data.written_memory_regions])

    return read_memory_regions, written_memory_regions


def get_path_entry(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns the first node in the successor path node_id belongs to."""
    current_node = node_id
    preds = get_predecessors(graph, current_node)
    while len(preds) != 0:
        current_node = preds[0]
        preds = get_predecessors(graph, current_node)
    return [current_node]


def get_parents(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns the first parent of node_id.
    Returns an empty list if none exists."""
    current_node = node_id
    preds = get_predecessors(graph, current_node)
    while len(preds) != 0:
        current_node = preds[0]
        preds = get_predecessors(graph, current_node)

    parents = [
        cast(int, edge[0]) for edge in graph.in_edges(current_node, data="data") if isinstance(edge[2], ChildEdge)
    ]

    return parents


def get_all_parents(graph: nx.DiGraph, node_id: int) -> List[int]:
    """Returns all parents of node_id.
    Returns an empty list if none exist."""

    queue: Set[int] = set(get_parents(graph, node_id))
    all_parents: Set[int] = set()
    while queue:
        current = queue.pop()
        all_parents.add(current)
        new_parents = [
            parent for parent in get_parents(graph, current) if parent not in queue and parent not in all_parents
        ]
        queue.update(new_parents)
    return list(all_parents)


global_graph = None
global_arguments = None


def get_available_decisions_for_functions(
    graph: nx.DiGraph, arguments: OptimizerArguments
) -> Dict[FunctionRoot, List[List[int]]]:
    """Returns a list of Lists describing all available decisions through the subgraph of the individual Functions."""
    global global_graph
    global global_arguments
    global_graph = graph
    global_arguments = arguments

    if arguments.verbose:
        print("Calculating available decisions per function...")
    available_decisions: Dict[FunctionRoot, List[List[int]]] = dict()
    param_list = [(function_node) for function_node in get_all_function_nodes(graph)]
    #    with Pool(
    #        initializer=__initialize_availability_worker,
    #        initargs=(graph, arguments),
    #    ) as pool:
    #        tmp_result = list(
    #            tqdm.tqdm(pool.imap_unordered(__parallel_get_decisions_from_node, param_list), total=len(param_list))
    #        )

    tmp_result: List[Any] = []
    for p in param_list:
        tmp_result.append(__parallel_get_decisions_from_node(p))

    for local_result in tmp_result:
        available_decisions[cast(FunctionRoot, data_at(graph, local_result[0]))] = local_result[1]

    if arguments.verbose:
        print("\tDone.")
        print("# Identified paths per function:")
        for function in available_decisions:
            print("#", function.name)
            for elem in available_decisions[function]:
                print("#..", elem)
        print()

    # show(graph, show_dataflow=False)
    # import sys
    # sys.exit(0)

    return available_decisions


def __initialize_availability_worker(graph: nx.DiGraph, arguments: OptimizerArguments):
    global global_graph
    global global_arguments
    global_graph = graph
    global_arguments = arguments


def __parallel_get_decisions_from_node_recursive(function_node):
    global global_graph
    global global_arguments

    def get_decisions_from_node_recursive(node_id, prev_decisions: List[int]) -> List[List[int]]:
        children_paths: List[List[int]] = []
        # get decisions from children
        for child in get_children(global_graph, node_id):
            children_paths += get_decisions_from_node_recursive(child, copy.deepcopy(prev_decisions))
        # get decisions from current node
        successors = get_successors(global_graph, node_id)
        successor_paths: List[List[int]] = []
        if len(successors) == 1:
            if len(children_paths) > 0:
                for cp in children_paths:
                    successor_paths += get_decisions_from_node_recursive(successors[0], copy.deepcopy(cp))
            else:
                successor_paths += get_decisions_from_node_recursive(successors[0], copy.deepcopy(prev_decisions))
            return successor_paths
        elif len(successors) == 0:
            if len(children_paths) > 0:
                return children_paths
            else:
                return [prev_decisions]
        else:
            tmp_result = []
            if len(children_paths) > 0:
                for cp in children_paths:
                    successor_paths = []
                    for succ in successors:
                        successor_paths += get_decisions_from_node_recursive(succ, copy.deepcopy(cp + [succ]))
                    tmp_result += successor_paths
            else:
                successor_paths = []
                for succ in successors:
                    successor_paths += get_decisions_from_node_recursive(succ, copy.deepcopy(prev_decisions + [succ]))
                tmp_result += successor_paths
            return tmp_result

    return function_node, get_decisions_from_node_recursive(function_node, [])


def __parallel_get_decisions_from_node(function_node):
    global global_graph
    global global_arguments

    # connected decisions in sets
    decision_sets: List[Set[int]] = []

    queue: List[int] = [function_node]
    visited: Set[int] = set()

    while len(queue) > 0:
        current = queue.pop(0)
        visited.add(current)
        queue += [c for c in get_children(global_graph, current) if c not in queue and c not in visited]
        queue += [s for s in get_successors(global_graph, current) if s not in queue and s not in visited]

        alternatives = set(get_out_mutex_edges(global_graph, current)).union(
            set(get_in_mutex_edges(global_graph, current))
        )
        if len(alternatives) == 0:
            continue
        # check if an existing decision set needs to be extended
        found_existing = False
        for idx, dcs in enumerate(decision_sets):
            if len(dcs.intersection(alternatives)) > 0:
                found_existing = True
                decision_sets[idx] = decision_sets[idx].union(alternatives)
        if not found_existing:
            decision_sets.append(alternatives)

    # convert sets to lists
    decision_lists: List[List[int]] = []
    for ds in decision_sets:
        decision_lists.append(list(ds))

    return function_node, decision_lists


#    def get_decisions_from_node(node_id, prev_decisions: List[int]) -> List[List[int]]:
#        children_paths: List[List[int]] = []
#        # get decisions from children
#        for child in get_children(global_graph, node_id):
#            children_paths += get_decisions_from_node(child, copy.deepcopy(prev_decisions))
#        # get decisions from current node
#        successors = get_successors(global_graph, node_id)
#        successor_paths: List[List[int]] = []
#        if len(successors) == 1:
#            if len(children_paths) > 0:
#                for cp in children_paths:
#                    successor_paths += get_decisions_from_node(successors[0], copy.deepcopy(cp))
#            else:
#                successor_paths += get_decisions_from_node(successors[0], copy.deepcopy(prev_decisions))
#            return successor_paths
#        elif len(successors) == 0:
#            if len(children_paths) > 0:
#                return children_paths
#            else:
#                return [prev_decisions]
#        else:
#            tmp_result = []
#            if len(children_paths) > 0:
#                for cp in children_paths:
#                    successor_paths = []
#                    for succ in successors:
#                        successor_paths += get_decisions_from_node(succ, copy.deepcopy(cp + [succ]))
#                    tmp_result += successor_paths
#            else:
#                successor_paths = []
#                for succ in successors:
#                    successor_paths += get_decisions_from_node(succ, copy.deepcopy(prev_decisions + [succ]))
#                tmp_result += successor_paths
#            return tmp_result

# return function_node, get_decisions_from_node(function_node, [])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Optional

from .PatternInfo import PatternInfo
from ..PEGraphX import (
    PEGraphX,
    Node,
    LineID,
    MemoryRegion,
)


class DeviceUpdateInfo(PatternInfo):
    """Class, that contains omp device updates"""

    def __init__(
        self,
        pet: PEGraphX,
        source_node: Node,
        target_node: Node,
        mem_reg: MemoryRegion,
        var_name: Optional[str],
        source_device_id: int,
        target_device_id: int,
        start_line: LineID,
        end_line: LineID,
        is_first_data_occurrence: bool,
        openmp_source_device_id: int,
        openmp_target_device_id: int,
    ):
        """
        :param pet: PET graph
        :param source_node: node, where data originates
        :param target_node: node, where data is required
        """
        PatternInfo.__init__(self, source_node)
        self.source_node_id = source_node.id
        self.target_node_id = target_node.id
        self.mem_reg = mem_reg
        self.var_name = var_name
        self.source_device_id = source_device_id
        self.target_device_id = target_device_id
        # overwrite start and end lines set by PatternInfo
        self.start_line = start_line
        self.end_line = end_line
        self.is_first_data_occurrence = is_first_data_occurrence
        self.openmp_source_device_id = openmp_source_device_id
        self.openmp_target_device_id = openmp_target_device_id

    def __str__(self):
        return (
            # f"Device update at: {self.node_id}\n" # removed to allow reduction of duplicates in the generated OpenMP code in CodeGenerator
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            # f"source_node: {self.source_node_id}\n"  # removed to allow reduction of duplicates in the generated OpenMP code in CodeGenerator
            # f"target_node: {self.target_node_id}\n"  # removed to allow reduction of duplicates in the generated OpenMP code in CodeGenerator
            f"source_device: {self.source_device_id}\n"
            f"target_device: {self.target_device_id}\n"
            # f"mem_reg: {self.mem_reg}\n"  # removed to allow reduction of duplicates in the generated OpenMP code in CodeGenerator
            f"var_name: {self.var_name}\n"
            f"is_first_data_occurrence: {self.is_first_data_occurrence}\n"
            f"openmp_source_device_id: {self.openmp_source_device_id}\n"
            f"openmp_target_device_id: {self.openmp_target_device_id}"
        )

    def get_str_without_first_data_occurrence(self) -> str:
        return (
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            f"source_device: {self.source_device_id}\n"
            f"target_device: {self.target_device_id}\n"
            f"var_name: {self.var_name}\n"
            f"openmp_source_device_id: {self.openmp_source_device_id}\n"
            f"openmp_target_device_id: {self.openmp_target_device_id}"
        )



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import cast
import warnings

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode


class ContextMerge(ContextNode):
    def __init__(self, node_id: int, experiment):
        super().__init__(node_id, experiment)

    def get_plot_label(self) -> str:
        return str(self.node_id) + "\nCTX\nmerge"

    def get_modified_context(
        self, node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject
    ) -> ContextObject:
        if len(context.snapshot_stack) < 1 or len(context.save_stack) < 1:
            warnings.warn("Context can not be merged before creating a snapshot!")
            return context

        context.last_seen_device_ids = context.snapshot_stack[-1][2]

        for saved_contex in context.save_stack[-1]:
            context.necessary_updates.update(saved_contex[1])
            for device_id in saved_contex[0]:
                if device_id not in context.seen_writes_by_device:
                    context.seen_writes_by_device[device_id] = dict()
                for mem_reg_id in saved_contex[0][device_id]:
                    if mem_reg_id not in context.get_seen_writes_by_device(device_id):
                        context.initialize_seen_writes_by_device(device_id, mem_reg_id)
                    if device_id not in context.seen_writes_by_device:
                        context.seen_writes_by_device[device_id] = dict()
                    if mem_reg_id not in context.seen_writes_by_device[device_id]:
                        context.seen_writes_by_device[device_id][mem_reg_id] = set()
                    context.seen_writes_by_device[device_id][mem_reg_id].update(saved_contex[0][device_id][mem_reg_id])

            context.last_seen_device_ids.append(saved_contex[2])
        return context



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os
import pathlib
import warnings
from typing import Any, Dict, List, Tuple, Optional

from sympy import Float, Symbol, Expr, Integer
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution

from discopop_library.discopop_optimizer.classes.system.Network import Network
from discopop_library.discopop_optimizer.classes.system.devices.CPU import CPU
from discopop_library.discopop_optimizer.classes.system.devices.Device import Device
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments


class System(object):
    __devices: Dict[int, Device]
    __host_device_id: int
    __network: Network
    __device_do_all_overhead_models: Dict[Device, Expr]
    __device_reduction_overhead_models: Dict[Device, Expr]
    __symbol_substitutions: List[
        Tuple[
            Symbol,
            Optional[float],
            Optional[float],
            Optional[float],
            Optional[FreeSymbolDistribution],
        ]
    ]

    def __init__(self, arguments: OptimizerArguments):
        self.__devices = dict()
        self.__host_device_id = -1
        self.__network = Network()
        self.__device_do_all_overhead_models = dict()
        self.__device_reduction_overhead_models = dict()
        self.__symbol_substitutions = []

        self.__build_from_configuration_file(arguments)

    # todo: support the replication of device ids (e.g. CPU-0 and GPU-0)

    def __build_from_configuration_file(self, arguments: OptimizerArguments):
        with open(arguments.system_configuration_path, "r") as f:
            system_configuration = json.load(f)
        self.__host_device_id = system_configuration["host_device"]
        # register devices
        for device_dict in system_configuration["devices"]:
            if device_dict["device_type"] == DeviceTypeEnum.CPU:
                self.__build_CPU(device_dict)
            elif device_dict["device_type"] == DeviceTypeEnum.GPU:
                self.__build_GPU(device_dict)
            else:
                raise ValueError("Unknown device type: " + str(device_dict["device_type"]) + " in: " + str(device_dict))
        # register host device for device -> device transfers
        self.__network.set_host_device(self.get_device(self.__host_device_id))

        # register connections
        for device_dict in system_configuration["devices"]:
            if "transfer_speeds" in device_dict:
                # register connection with given transfer speed
                self.__network.add_connection(
                    source=self.get_device(self.__host_device_id),
                    target=self.get_device(device_dict["device_id"]),
                    transfer_speed=Float(device_dict["transfer_speeds"]["H2D_MB/s"]),
                    initialization_delay=Float(device_dict["transfer_init_delays[us]"]["average"]),
                )  # H2D
                self.__network.add_connection(
                    source=self.get_device(device_dict["device_id"]),
                    target=self.get_device(self.__host_device_id),
                    transfer_speed=Float(device_dict["transfer_speeds"]["D2H_MB/s"]),
                    initialization_delay=Float(device_dict["transfer_init_delays[us]"]["average"]),
                )  # D2H
            else:
                # no transfer speed information exists
                pass

    def __build_CPU(self, device_configuration: Dict[str, Any]):
        cpu = CPU(
            frequency=Integer(device_configuration["frequency"]),
            thread_count=Integer(device_configuration["threads"]),
            openmp_device_id=device_configuration["device_id"],
            device_specific_compiler_flags="",
        )
        self.add_device(cpu, device_configuration["device_id"])

    def __build_GPU(self, device_configuration: Dict[str, Any]):
        gpu = GPU(
            frequency=Integer(device_configuration["frequency"]),
            thread_count=Integer(device_configuration["threads"]),
            openmp_device_id=device_configuration["device_id"],
            device_specific_compiler_flags="",
        )
        self.add_device(gpu, device_configuration["device_id"])

    def set_device_doall_overhead_model(self, device: Device, model: Expr, arguments: OptimizerArguments):
        if arguments.verbose:
            print("System: Set DOALL overhead model: ", model)
        self.__device_do_all_overhead_models[device] = model

    def set_reduction_overhead_model(self, device: Device, model: Expr, arguments: OptimizerArguments):
        if arguments.verbose:
            print("System: Set REDUCTION overhead model: ", model)
        self.__device_reduction_overhead_models[device] = model

    def get_device_doall_overhead_model(self, device: Device, arguments: OptimizerArguments) -> Expr:
        if device not in self.__device_do_all_overhead_models:
            if arguments.verbose:
                warnings.warn("No DOALL overhead model, assuming 0 for device: " + str(device))
            return Expr(Integer(0))
        return self.__device_do_all_overhead_models[device]

    def get_device_reduction_overhead_model(self, device: Device, arguments: OptimizerArguments) -> Expr:
        if device not in self.__device_reduction_overhead_models:
            if arguments.verbose:
                warnings.warn("No REDUCTION overhead model, assuming 0 for device: " + str(device))
            return Expr(Integer(0))
        return self.__device_reduction_overhead_models[device]

    def add_device(self, device: Device, device_id: int):
        self.__devices[device_id] = device

    def get_device(self, device_id: Optional[int]) -> Device:
        if device_id is None:
            return self.__devices[self.get_host_device_id()]
        return self.__devices[device_id]

    def get_device_ids_by_type(self, device_type: type) -> List[int]:
        result_device_ids = []
        for device_id in self.__devices:
            if type(self.__devices[device_id]) == device_type:
                result_device_ids.append(device_id)
        return result_device_ids

    def get_network(self) -> Network:
        return self.__network

    def get_free_symbols(self) -> List[Tuple[Symbol, Optional[Expr]]]:
        result_list: List[Tuple[Symbol, Optional[Expr]]] = []
        for device in self.__devices.values():
            result_list += device.get_free_symbols()
        result_list += self.__network.get_free_symbols()
        return result_list

    def get_device_id(self, device: Device) -> int:
        for key in self.__devices:
            if device == self.__devices[key]:
                return key
        raise ValueError("Unknown device: ", device)

    def get_host_device_id(self) -> int:
        return self.__host_device_id

    def get_symbol_values_and_distributions(
        self,
    ) -> List[Tuple[Symbol, Optional[float], Optional[float], Optional[float], Optional[FreeSymbolDistribution]]]:
        return self.__symbol_substitutions


# define a default system



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
from dataclasses import dataclass


@dataclass
class PatchGeneratorArguments(object):
    """Container Class for the arguments passed to the discopop_patch_generator"""

    verbose: bool
    discopop_build_path: str
    CC: str
    CXX: str
    add_from_json: str

    def __post_init__(self):
        self.__validate()

    def __validate(self):
        """Validate the arguments passed to the discopop_explorer, e.g check if given files exist"""
        if self.verbose:
            print("Configuration:")
            print("\tDP BUILD PATH: ", self.discopop_build_path)
            print("\tCC: ", self.CC)
            print("\tCXX: ", self.CXX)
        # check if build directory exists
        if not os.path.exists(self.discopop_build_path):
            raise FileNotFoundError(self.discopop_build_path)
        # check if CC and CXX exist
        if not os.path.exists(self.CC):
            raise FileNotFoundError(self.CC)
        if not os.path.exists(self.CXX):
            raise FileNotFoundError(self.CXX)

        pass



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Set, List, Tuple

from discopop_explorer.PEGraphX import PEGraphX, NodeID
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import VarName
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Dependency import Dependency


def get_dependencies_as_metadata(
    pet: PEGraphX, all_dependencies: Set[Dependency]
) -> Tuple[List[Tuple[VarName, NodeID, str]], List[Tuple[VarName, NodeID, str]],]:
    in_deps_metadata: Set[Tuple[VarName, NodeID, str]] = set()
    out_deps_metadata: Set[Tuple[VarName, NodeID, str]] = set()

    for dependency in all_dependencies:
        # create out dependency metadata
        for var_name in dependency.var_names:
            out_deps_metadata.add(
                (
                    var_name,
                    dependency.source,
                    pet.node_at(dependency.source).end_position(),
                )
            )

        # create in dependency metadata
        for var_name in dependency.var_names:
            in_deps_metadata.add(
                (
                    var_name,
                    dependency.sink,
                    pet.node_at(dependency.sink).start_position(),
                )
            )

    return list(in_deps_metadata), list(out_deps_metadata)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os

from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.UpdateOptimization.LoopInitializationUpdates import (
    fix_loop_initialization_updates,
)
from discopop_library.discopop_optimizer.UpdateOptimization.RemoveDuplicatedUpdates import remove_duplicated_updates
from discopop_library.discopop_optimizer.UpdateOptimization.RemoveLoopIndexUpdates import remove_loop_index_updates
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.utilities.visualization.update_graph import show_update_graph
from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern


def optimize_updates(
    experiment: Experiment, best_configuration: OptimizerOutputPattern, arguments: OptimizerArguments
) -> OptimizerOutputPattern:
    # plot raw update graph
    # show_update_graph(experiment.optimization_graph, best_configuration, experiment)

    # print original updates
    if arguments.verbose:
        print("Original updates")
        for update in best_configuration.data_movement:
            print("# ", update)
        print()

    # optimize updates
    best_configuration = fix_loop_initialization_updates(experiment, best_configuration, arguments)

    # remove duplicated updates
    best_configuration = remove_duplicated_updates(best_configuration, arguments)

    # remove loop index updates
    best_configuration = remove_loop_index_updates(experiment, best_configuration, arguments)

    # plt optimized update graph
    # show_update_graph(experiment.optimization_graph, best_configuration, experiment)

    # print optimized updates
    if arguments.verbose:
        print("Optimized updates")
        for update in best_configuration.data_movement:
            print("# ", update)
        print()

    return best_configuration


#    # export the updated configuration to the disk
#    updated_configuration_path = os.path.join("optimizer", "updated_configuration.json")
#    best_configuration.dump_to_file(updated_configuration_path)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import List, Optional, Set

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.utilities.MOGUtilities import get_parents


def is_child_of_any(graph: nx.DiGraph, start_node: int, potential_parents: List[Optional[int]]) -> bool:
    """returns True, if start_node is a child of any of the potential_parents"""
    # get all parents of start_node
    all_parents: Set[int] = set()
    cur_parents = get_parents(graph, start_node)
    while len(cur_parents) != 0:
        cp = cur_parents.pop(0)
        all_parents.add(cp)
        cur_parents += get_parents(graph, cp)

    # check if an intersection exists with potential parents
    intersection = all_parents.intersection(set(potential_parents))

    # if so, start_node is a child of the potential parents
    if len(intersection) != 0:
        return True
    return False



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from discopop_explorer.pattern_detectors.simple_gpu_patterns.GPULoop import GPULoopPattern


def sort_by_nodeID(e: GPULoopPattern):
    """used to sort a list of gpu patterns by their node ids

    :return:
    :param e:
    """
    return e.nodeID
